from django.core.management.base import BaseCommand  # , CommandError
from django.contrib.auth import get_user_model

User = get_user_model()


class Command(BaseCommand):
    help = 'Sets apigw_consumer_id to "None"'

    def handle(self, *args, **options):
        users = User.objects.filter(is_staff=False, is_superuser=False)
        users_count = users.count()
        updated_users = 0

        for user in users:
            try:
                profile = user.sso_app_profile
                profile.apigw_consumer_id = None
                profile.save()

                updated_users += 1
                self.stdout.write(self.style.SUCCESS('{}/{}'.format(updated_users, users_count)))

            except Exception as e:
                self.stdout.write(self.style.ERROR('Error "{}" updating apigw_consumer_id for "{}"'.format(e, user)))

        self.stdout.write(self.style.SUCCESS('Updated {}/{} users'.format(updated_users, users_count)))



import logging
# import inspect

from django.contrib import auth
from django.urls import reverse
from django.conf import settings

from ...permissions import is_authenticated
from ...exceptions import ServiceSubscriptionRequiredException, ProfileIncompleteException
from ...utils import check_user_can_login, set_session_key, get_session_key
from ... import app_settings

from .base import DjangoSsoAppAuthenticationBaseMiddleware

logger = logging.getLogger('django_sso_app.core.middleware')

ADMIN_URL = '/{}'.format(settings.ADMIN_URL)
SSO_ID_JWT_KEY = 'sso_id'
PROFILE_INCOMPLETE_ENABLED_PATHS = [
    reverse('javascript-catalog'),
    reverse('profile.complete'),
]


class DjangoSsoAppAuthenticationAppMiddleware(DjangoSsoAppAuthenticationBaseMiddleware):

    def app_process_request(self, request, request_jwt, decoded_jwt):
        user = getattr(request, 'user', None)

        if not is_authenticated(user):
            # We are seeing this user for the first time in this session, attempt
            # to authenticate the user.
            try:
                if app_settings.APIGATEWAY_ENABLED:
                    consumer_username = get_session_key(request, '__dssoa__apigateway__consumer_username')

                    user = auth.authenticate(request=request,
                                             consumer_username=consumer_username,
                                             encoded_jwt=request_jwt,
                                             decoded_jwt=decoded_jwt)
                else:
                    user = auth.authenticate(request=request,
                                             encoded_jwt=request_jwt,
                                             decoded_jwt=decoded_jwt)

            except ServiceSubscriptionRequiredException as e:
                logger.debug('ServiceSubscriptionRequiredException. Redirecting to service subscription')

                set_session_key(request, '__dssoa__redirect', e.response)

            except ProfileIncompleteException as e:
                logger.debug('ProfileIncompleteException. Redirecting to profile completion')

                set_session_key(request, '__dssoa__redirect', e.response)

            except Exception as e:
                logger.exception('App middleware exception "{}"'.format(e))

                self._remove_invalid_user(request)

                return

            else:
                logger.info('User "{}" authenticated successfully!'.format(user))

                # set request user
                setattr(request, 'user', user)  # !!

        # further checks
        if is_authenticated(user):
            set_session_key(request, '__dssoa__requesting_user', user)

    # response
    def app_process_response(self, request, response):

        redirect = get_session_key(request, '__dssoa__redirect', None)

        if redirect is not None:
            return redirect

        # redirect to profile complete (if required)
        if get_session_key(request, '__dssoa__must_update_profile', False):
            requesting_user = get_session_key(request, '__dssoa__requesting_user')

            logger.info('Redirecting user "{}" to profile completion'.format(requesting_user))

            response = self._redirect_to_profile_complete_response(request.user)

        return response



import logging

from allauth.account.models import (
    EmailAddress,
)
from django.contrib.auth import get_user_model
from django.core import mail
from django.urls import reverse
from django.test.utils import override_settings

from rest_framework import status

from allauth.account.adapter import get_adapter

from .factories import UserTestCase

User = get_user_model()
logger = logging.getLogger('django_sso_app.core.tests')


class TestLogin(UserTestCase):
    def test_can_login_with_email(self):
        new_pass = self._get_random_pass()
        new_user = self._get_new_user(password=new_pass)

        client = self._get_client()
        response = client.post(
            reverse('account_login'),
            data=self._get_login_object(new_user.email, new_pass)
        )

        self.assertEqual(response.status_code, status.HTTP_302_FOUND)
        self.assertEqual(response.url, '/')
        self.assertIsNotNone(self._get_response_jwt_cookie(response), 'response has no jwt cookie')

    def test_can_login_with_username(self):
        new_pass = self._get_random_pass()
        new_user = self._get_new_user(password=new_pass)

        client = self._get_client()
        response = client.post(
            reverse('account_login'),
            data=self._get_login_object(new_user.username, new_pass)
        )

        self.assertEqual(response.status_code, status.HTTP_302_FOUND)
        self.assertEqual(response.url, '/')
        self.assertIsNotNone(self._get_response_jwt_cookie(response), 'response has no jwt cookie')

    def test_admin_user_do_not_receives_jwt_on_login(self):
        new_admin_pass = self._get_random_pass()
        new_admin_user = self._get_new_admin_user(password=new_admin_pass)

        response = self.client.post(
            '/admin/login/',
            data=self._get_login_object(new_admin_user.username, new_admin_pass)
        )

        self.assertIsNone(self._get_response_jwt_cookie(response), 'admin receives jwt')

    def test_django_superuser_can_not_login_from_default_views(self):
        new_admin_pass = self._get_random_pass()
        new_admin_user = self._get_new_admin_user(password=new_admin_pass)

        client = self._get_client()
        response = client.post(
            reverse('account_login'),
            data=self._get_login_object(new_admin_user.username, new_admin_pass)
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIsNone(self._get_response_jwt_cookie(response), 'admin receives jwt')

    def test_django_staff_user_can_not_login_from_default_views(self):
        new_staff_pass = self._get_random_pass()
        new_staff_user = self._get_new_staff_user(password=new_staff_pass)

        client = self._get_client()
        response = client.post(
            reverse('account_login'),
            data=self._get_login_object(new_staff_user.username, new_staff_pass)
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIsNone(self._get_response_jwt_cookie(response), 'staff receives jwt')

    def test_user_must_complete_profile(self):
        new_pass = self._get_random_pass()
        new_user = self._get_new_incomplete_user(password=new_pass)

        client = self._get_client()
        response = client.post(
            reverse('account_login'),
            data=self._get_login_object(new_user.email, new_pass),
            follow=True
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self._test_response_path(response, reverse('profile.complete'))
        self.assertIsNotNone(self._get_response_jwt_cookie(response), 'incomplete user do not receives jwt')


class TestLogout(UserTestCase):

    @override_settings(DJANGO_SSO_APP_LOGOUT_DELETES_ALL_PROFILE_DEVICES=True)
    def test_logout_deletes_jwt_and_removes_all_profile_devices(self):
        new_pass = self._get_random_pass()
        new_user = self._get_new_user(password=new_pass)

        client = self._get_client()
        response = client.post(
            reverse('account_login'),
            data=self._get_login_object(new_user.email, new_pass)
        )

        self.assertEqual(response.status_code, status.HTTP_302_FOUND)

        response2 = client.post(
            reverse('account_logout')
        )

        self.assertEqual(self._get_response_jwt_cookie(response2), '', 'api logout keeps jwt cookie')

        user_devices_post_logout = new_user.sso_app_profile.devices.all()

        self.assertEqual(user_devices_post_logout.count(), 0,
                         'api logout is not deleting all profile devices')

    @override_settings(DJANGO_SSO_APP_LOGOUT_DELETES_ALL_PROFILE_DEVICES=False)
    def test_logout_deletes_jwt_and_removes_caller_device(self):
        new_pass = self._get_random_pass()
        new_user = self._get_new_user(password=new_pass)

        client = self._get_client()
        response = client.post(
            reverse('account_login'),
            data=self._get_login_object(new_user.email, new_pass)
        )

        self.assertEqual(response.status_code, status.HTTP_302_FOUND)

        user_devices_pre_logout = new_user.sso_app_profile.devices.all()

        response2 = client.post(
            reverse('account_logout')
        )

        self.assertEqual(self._get_response_jwt_cookie(response2), '', 'api logout keeps jwt cookie')

        user_devices_post_logout = new_user.sso_app_profile.devices.all()

        self.assertEqual(user_devices_post_logout.count(), user_devices_pre_logout.count(),
                         'api logout is deleting all profile devices')


class TestSignup(UserTestCase):
    def test_can_register(self):
        signup_obj = self._get_signup_object()

        client = self._get_client()
        response = client.post(
            reverse('account_signup'),
            data=signup_obj
        )

        self.assertEqual(response.status_code, status.HTTP_302_FOUND)

        new_user = User.objects.filter(email=signup_obj['email']).first()

        self.assertIsNotNone(new_user)
        self.assertEqual(new_user.email, signup_obj['email'])

        email_address = EmailAddress.objects.filter(user=new_user, email=signup_obj['email']).first()

        self.assertNotEqual(email_address, None)

    def test_cannot_register_on_invalid_registration_data(self):
        users_count_before = User.objects.count()

        signup_obj = self._get_invalid_signup_object()

        client = self._get_client()
        response = client.post(
            reverse('account_signup'),
            data=signup_obj
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

        users_count_after = User.objects.count()

        self.assertEqual(users_count_before, users_count_after)

    def test_registration_sends_confirmation_email(self):
        signup_obj = self._get_signup_object()

        client = self._get_client()
        response = client.post(
            reverse('account_signup'),
            data=signup_obj
        )

        self.assertEqual(response.status_code, status.HTTP_302_FOUND)

        self.assertEqual(len(mail.outbox), 1)

    def test_registration_do_not_update_sso_rev(self):
        signup_obj = self._get_signup_object()

        client = self._get_client()
        response = client.post(
            reverse('account_signup'),
            data=signup_obj
        )

        self.assertEqual(response.status_code, status.HTTP_302_FOUND)

        new_user = User.objects.get(email=signup_obj['email'])

        self.assertEqual(new_user.sso_app_profile.sso_rev, 1)

    def test_registration_creates_subscription(self):
        new_service = self._get_new_service()
        signup_obj = self._get_subscribe_signup_object(new_service.service_url)

        client = self._get_client()
        response = client.post(
            reverse('account_signup'),
            data=signup_obj
        )

        self.assertEqual(response.status_code, status.HTTP_302_FOUND)

        new_user = User.objects.filter(email=signup_obj['email']).first()
        self.assertNotEqual(new_user, None, 'no new user')

        self.assertEqual(new_user.sso_app_profile.subscriptions.filter(service__service_url=
                                                                       new_service.service_url).count(), 1)

    def test_created_user_incomplete_profile_has_username_set_to_email(self):
        signup_obj = self._get_signup_object()

        client = self._get_client()
        response = client.post(
            reverse('account_signup'),
            data=signup_obj
        )

        self.assertEqual(response.status_code, status.HTTP_302_FOUND)

        new_user = User.objects.filter(username=signup_obj['email']).first()
        self.assertNotEqual(new_user, None, 'no new user with username as email')

    def test_created_user_with_incomplete_profile_enters_incomplete_group(self):
        signup_obj = self._get_signup_object()

        client = self._get_client()
        response = client.post(
            reverse('account_signup'),
            data=signup_obj
        )

        self.assertEqual(response.status_code, status.HTTP_302_FOUND)

        new_user = User.objects.filter(email=signup_obj['email']).first()
        self.assertNotEqual(new_user, None, 'no new incomplete user')

        unconfirmed_email = new_user.emailaddress_set.get(email=signup_obj['email'])
        self.assertEqual(unconfirmed_email.verified, False)

        adapter = get_adapter()
        adapter.confirm_email(None, unconfirmed_email)

        new_pass = signup_obj['password1']
        client = self._get_client()
        response = client.post(
            reverse('account_login'),
            data=self._get_login_object(new_user.email, new_pass),
            follow=True
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self._test_response_path(response, reverse('profile.complete'))
        self.assertIsNotNone(self._get_response_jwt_cookie(response), 'incomplete user do not receives jwt')

        self.assertIsNotNone(new_user.sso_app_profile.groups.filter(name='incomplete').first(),
                             'incomplete user did not enter "incomplete" group on login')



from django.core.management.base import BaseCommand  # , CommandError
from django.contrib.auth import get_user_model

from ...utils import get_profile_apigw_consumer_id  # , delete_apigw_consumer

User = get_user_model()


class Command(BaseCommand):
    help = 'Restores apigateway consumer by username'

    def add_arguments(self, parser):
        parser.add_argument('username', nargs='+', type=str)

    def handle(self, *args, **options):
        # for poll_id in options['poll_ids']:
        #     try:
        #         poll = Poll.objects.get(pk=poll_id)
        #     except Poll.DoesNotExist:
        #         raise CommandError('Poll "%s" does not exist' % poll_id)
        usernames = options['username']
        users = User.objects.filter(username__in=usernames)
        users_count = users.count()
        updated_users = 0

        for username in usernames:
            try:
                user = User.objects.get(username=username)
            except User.DoesNotExist:
                self.stdout.write(self.style.ERROR('User "{}" does not exist'.format(username)))
                continue

            try:
                consumer_id = get_profile_apigw_consumer_id(user.sso_app_profile, True)
                updated_users += 1
                self.stdout.write(self.style.SUCCESS('{}/{} "{}"'.format(updated_users, users_count, consumer_id)))

            except Exception as e:
                self.stdout.write(self.style.ERROR('Error "{}" restoring apigateway consumer for "{}"'.format(e, user)))

                # delete_apigw_consumer(user.sso_app_profile)

        self.stdout.write(self.style.SUCCESS('Restored {}/{} users'.format(updated_users, users_count)))



from rest_framework import serializers


class PartialObjectSerializer(serializers.Serializer):
    _partial = serializers.SerializerMethodField(method_name='get_partial')

    def get_partial(self, obj):
        return True


class AbsoluteUrlSerializer(serializers.ModelSerializer):
    url = serializers.SerializerMethodField(method_name='get_absolute_rest_url')

    def get_absolute_rest_url(self, obj):
        if getattr(obj, 'pk', None) is not None:
            get_absolute_rest_url = getattr(obj, 'get_absolute_rest_url', None)
            if get_absolute_rest_url is None:
                raise NotImplementedError('Model must provide "get_absolute_rest_url" method.')
            else:
                request = self.context['request']
                return request.build_absolute_uri(get_absolute_rest_url())


class UserRelatedSerializer(serializers.Serializer):
    user = serializers.SerializerMethodField()

    def get_user(self, obj):
        user = getattr(obj, 'user', None)
        if user is not None:
            profile = getattr(user, 'sso_app_profile', None)
            if profile is not None:
                request = self.context['request']
                reverse_url = profile.get_absolute_rest_url()
                return request.build_absolute_uri(reverse_url)


class DeactivableSerializer(serializers.Serializer):
    is_active = serializers.BooleanField(required=False)


class PublicableSerializer(serializers.Serializer):
    is_public = serializers.BooleanField(required=False)


class TimespanSerializer(serializers.Serializer):
    started_at = serializers.DateTimeField(required=False)
    ended_at = serializers.DateTimeField(required=False)


class UpdatableSerializer(serializers.Serializer):
    updated_at = serializers.DateTimeField(required=False)


class CreatedAtSerializer(serializers.Serializer):
    created_at = serializers.DateTimeField(required=False)



import logging

from django.dispatch import receiver
from django.contrib.auth.signals import user_logged_in
from django.db.models.signals import pre_save, post_save

from allauth.account.adapter import get_adapter

from ....utils import set_session_key
from ....tokens.utils import jwt_encode
from ....functions import get_random_string
from ....permissions import is_django_staff
from ...profiles.models import Profile

from .... import app_settings

from ..models import Device
from ..utils import get_request_device

logger = logging.getLogger('django_sso_app.core.apps.devices.signals')


# profile

@receiver(post_save, sender=Profile)
def post_profile_saved(sender, instance, created, **kwargs):
    if kwargs['raw']:
        # https://github.com/django/django/commit/18a2fb19074ce6789639b62710c279a711dabf97
        return

    profile = instance
    user = instance.user

    if not is_django_staff(user):
        if not created:  # if instance.pk:
            logger.debug('Profile "{}" post_save signal'.format(profile))

            rev_updated = getattr(profile, '__rev_updated', False)

            if rev_updated:
                logger.info('Rev updated, removing all user devices for Profile "{}"'.format(profile))
                adapter = get_adapter()

                adapter.remove_all_user_profile_devices(profile.user)


# device

@receiver(pre_save, sender=Device)
def pre_device_save(sender, instance, **kwargs):
    if not app_settings.APIGATEWAY_ENABLED:
        user = instance.profile.user
        device = instance

        if not is_django_staff(user):
            logger.debug('Skip creating api gateway JWT for Device "{}"'.format(device))

            device.apigw_jwt_id = None
            device.apigw_jwt_key = None
            device.apigw_jwt_secret = device.apigw_jwt_secret or get_random_string(32)


@receiver(user_logged_in)
def post_user_login(**kwargs):
    user = kwargs['user']

    if not is_django_staff(user):
        request = kwargs['request']

        device = get_request_device(request)
        token = jwt_encode(device.get_jwt_payload(), device.apigw_jwt_secret)

        set_session_key(request, '__dssoa__device', device)
        set_session_key(request, '__dssoa__jwt_token', token)

        logger.info('User Logged in, request has device "{}"'.format(device))



import base64
from io import BytesIO

from PIL import Image
from django import forms
from django.contrib import admin
from django.core.files.uploadedfile import InMemoryUploadedFile, TemporaryUploadedFile

from .models import Service, Subscription, TOS


class ServiceAdminForm(forms.ModelForm):
    picture = forms.ImageField(required=False)

    def clean_picture(self):
        picture_file = self.cleaned_data['picture']

        if picture_file is not None and type(picture_file) in [InMemoryUploadedFile, TemporaryUploadedFile]:
            picture_bin = picture_file.file.read()
            picture_image = Image.open(BytesIO(picture_bin))
            picture_format = picture_image.format

            if picture_format not in ('PNG', 'JPEG'):
                raise forms.ValidationError('Image must be either PNG or JPEG')

            picture_image.thumbnail((100, 100), Image.ANTIALIAS)
            
            image = BytesIO()
            picture_image.save(image, format=picture_format)

            return "data:image/{};base64,".format(picture_format.lower()) + base64.b64encode(image.getvalue()).decode('utf-8')
        else:
            return None


class TOSInline(admin.TabularInline):
    model = TOS


class ServiceAdmin(admin.ModelAdmin):
    form = ServiceAdminForm
    inlines = [
        TOSInline,
    ]


class SubscriptionAdmin(admin.ModelAdmin):
    search_fields = ('id', 'profile__user__username', 'profile__sso_id')
    list_display = ('service', 'profile', 'created_at', 'unsubscribed_at')


admin.site.register(Service, ServiceAdmin)
admin.site.register(Subscription, SubscriptionAdmin)



import logging

from django.db.models.signals import m2m_changed, post_save
from django.dispatch import receiver
from django.contrib.auth.signals import user_logged_in

from .... import app_settings
from ....permissions import is_django_staff
from ...profiles.models import Profile
from ...users.utils import add_user_to_default_groups
from ..models import Group

logger = logging.getLogger('django_sso_app.core.apps.groups.signals')


@receiver(m2m_changed, sender=Profile.groups.through)
def signal_handler_when_user_profile_is_added_or_removed_from_group(action, instance, pk_set, **kwargs):
    """
    Update sso_rev when user enter/exit groups
    """

    profile = instance
    user = profile.user

    logger.debug('Groups signal')

    is_loaddata = getattr(user, '__dssoa__loaddata', False)
    is_creating = getattr(user, '__dssoa__creating', False)
    is_entering_incomplete_group = getattr(user, '__dssoa__enter_incomplete_group', False)
    is_exiting_incomplete_group = getattr(user, '__dssoa__exit_incomplete_group', False)

    must_update_rev = (not is_loaddata) and \
                      (not is_creating) and \
                      (not is_entering_incomplete_group) and \
                      (not is_exiting_incomplete_group) and \
                      (not is_django_staff(user))

    groups_updated = False

    if action == 'pre_add':
        groups_updated = True
        for pk in pk_set:
            group = Group.objects.get(id=pk)
            logger.info('Profile "{}" entered group "{}"'.format(profile, group))

    elif action == 'pre_remove':
        groups_updated = True
        for pk in pk_set:
            group = Group.objects.get(id=pk)
            logger.info('Profile "{}" exited from group "{}"'.format(profile, group))

    # updating rev

    if groups_updated and must_update_rev:
        profile.update_rev(True)


@receiver(post_save, sender=Profile)
def user_profile_updated(sender, instance, created, **kwargs):
    """
    Profile model has been updated, check user profile completeness
    """
    profile = instance
    user = profile.user

    if kwargs['raw']:
        # https://github.com/django/django/commit/18a2fb19074ce6789639b62710c279a711dabf97
        return

    if not is_django_staff(user):
        add_user_to_default_groups(user)

        should_manage_groups = app_settings.BACKEND_ENABLED or app_settings.REPLICATE_PROFILE

        if should_manage_groups:
            if created:  # if instance.pk:
                if profile.is_incomplete:
                    logger.debug('Profile model has been created incomplete, entering "incomplete" group')

                    setattr(profile.user, '__dssoa__enter_incomplete_group', True)

                    group, _created = Group.objects.get_or_create(name='incomplete')

                    profile.groups.add(group)
            else:
                if not profile.is_incomplete:
                    group, _created = Group.objects.get_or_create(name='incomplete')

                    if profile.groups.filter(name=group.name).count() > 0:
                        logger.debug('Profile model has been completed, exiting "incomplete" group')

                        setattr(profile.user, '__dssoa__exit_incomplete_group', True)

                        profile.groups.remove(group)

            # check profile groups
            user_object_profile = getattr(profile.user, '__dssoa__profile__object', None)

            if user_object_profile is not None:
                group_names = user_object_profile.get('groups', [])

                for group_name in group_names:
                    group, _created = Group.objects.get_or_create(name=group_name)

                    if profile.groups.filter(name=group.name).count() == 0:
                        profile.groups.add(group)


@receiver(user_logged_in)
def post_user_login(**kwargs):
    """
    Check if user belongs to all DEFAULT_USER_GROUPS, if not
    """
    user = kwargs['user']

    if not is_django_staff(user):
        add_user_to_default_groups(user)



# Boostable functions
#
# import pyximport
# pyximport.install()


def get_apigateway_consumer_id(sso_id: str) -> str:
    """
    Translates from kong consumer.username to sso_id
    :param sso_id:
    :return:
    """
    return str(sso_id).replace('-', '_')


def get_apigateway_sso_id(apigateway_user_id: str) -> str:
    """
    Translates from sso_id to kong consumer.username
    :param apigateway_user_id:
    :return:
    """
    return str(apigateway_user_id).replace('_', '-')


def get_apigateway_profile_groups_from_header(groups_header: str) -> list:
    """
    Extracts apigateway consumer groups from header
    :param groups_header:
    :return:
    """
    if groups_header not in [None, '']:
        return list(map(str.strip, groups_header.split(',')))

    return []



from django.core.management.base import BaseCommand, CommandError
from django.contrib.auth import get_user_model

from allauth.account.adapter import get_adapter

User = get_user_model()


class Command(BaseCommand):
    help = 'Deletes all profiles devices'

    def handle(self, *args, **options):
        users = User.objects.all()
        users_count = users.count()
        updated_users = 0
        parsed_users = 0
        devices_count = 0
        deleted_devices = 0
        adapter = get_adapter()

        for user in users:
            try:
                profile_devices_count = user.sso_app_profile.devices.count()
                devices_count += profile_devices_count
                deleted_profile_devices_count = adapter.remove_all_user_profile_devices(user)

                if deleted_profile_devices_count > 0:
                    updated_users += 1
                    deleted_devices += deleted_profile_devices_count

                    self.stdout.write(
                        self.style.SUCCESS('Deleted {}/{} user devices for "{}"'.format(profile_devices_count,
                                                                                        deleted_profile_devices_count,
                                                                                        user)))

            except Exception as e:
                self.stdout.write(self.style.ERROR('Error "{}" deleting devices for "{}"'.format(e, user)))

            parsed_users += 1
            self.stdout.write(self.style.SUCCESS('{}/{}'.format(parsed_users, users_count)))

        self.stdout.write(self.style.SUCCESS('Updated {}/{} profiles'.format(users_count, updated_users)))
        self.stdout.write(self.style.SUCCESS('Deleted {}/{} devices'.format(deleted_devices, devices_count)))



import logging

from django.contrib.messages import get_messages
from django.utils.encoding import force_str
from django.shortcuts import reverse

logger = logging.getLogger('django_sso_app.core.api')


def get_request_messages_string(request):
    """
    Serializes django messages

    :param request:
    :return:
    """
    storage = get_messages(request)
    _messages = []
    for message in storage:
        _messages.append(force_str(message))

    return ', '.join(_messages)



from django.conf.urls import url
from rest_framework.urlpatterns import format_suffix_patterns

from .views import GroupApiViewSet


urlpatterns = [
    url(r'^$', GroupApiViewSet.as_view({'get': 'list'}), name="rest-list"),
    url(r'^(?P<pk>\w+)/$', GroupApiViewSet.as_view({'get': 'retrieve'}), name="rest-detail"),
]

urlpatterns = (format_suffix_patterns(urlpatterns), 'django_sso_app_group')



from django.http import HttpResponseRedirect
from django.views import View

from ..core import app_settings


class AppLoginView(View):

    def get(self, request, *args, **kwargs):
        return HttpResponseRedirect(app_settings.REMOTE_LOGIN_URL)


class AppSignupView(View):

    def get(self, request, *args, **kwargs):
        return HttpResponseRedirect(app_settings.REMOTE_SIGNUP_URL)


class AppLogoutView(View):

    def get(self, request, *args, **kwargs):
        return HttpResponseRedirect(app_settings.REMOTE_LOGOUT_URL)



import logging

from django.db.models.signals import post_save
from django.dispatch import receiver

from ..models import Subscription

logger = logging.getLogger('django_sso_app.core.apps.services.signals')


@receiver(post_save, sender=Subscription)
def check_upated_fields(sender, instance, created, **kwargs):
    if kwargs['raw']:
        # https://github.com/django/django/commit/18a2fb19074ce6789639b62710c279a711dabf97
        return

    if created:  # not instance._state.adding:  # if instance.pk:
        logger.info('Subsciption created, doing nothing')

    else:
        subscription = instance

        if subscription.is_unsubscribed or not subscription.is_active:
            user = subscription.profile.user

            if not getattr(user, '__dssoa__creating', False):
                logger.info('Subsciption updates profile rev for "{}"'.format(user))

                subscription.profile.update_rev(True)



import logging

from allauth.account.adapter import get_adapter

from ...tokens.utils import jwt_decode
from ...utils import get_session_key, set_session_key
from ...tokens.utils import get_request_jwt
from ...exceptions import RequestHasValidJwtWithNoDeviceAssociated

logger = logging.getLogger('django_sso_app.core.apps.devices')


def get_request_device(request):
    device = get_session_key(request, '__dssoa__device', None)

    if device is None:
        logger.debug('no previous device')
        fingerprint = get_session_key(request, '__dssoa__device__fingerprint')

        if fingerprint is None:
            logger.debug('received empty fingerprint')

            raw_token = get_request_jwt(request)

            if raw_token is None:
                fingerprint = 'undefined'
            else:
                try:
                    device, verified_payload = jwt_decode(raw_token, verify=True)

                except RequestHasValidJwtWithNoDeviceAssociated:
                    logger.warning('no device associated to request token "{}"'.format(raw_token))
                    raise
                else:
                    fingerprint = verified_payload['fp']

        if device is None:
            user = request.user
            profile = user.sso_app_profile

            logger.info(
                'Request has no device, getting profile "{}" device with fingerprint "{}"'.format(profile,
                                                                                                  fingerprint))

            device = profile.devices.filter(fingerprint=fingerprint).first()

            if device is None:
                logger.debug('profile has no device with fingerprint "{}"'.format(fingerprint))
                adapter = get_adapter(request)
                device = adapter.add_user_profile_device(user, fingerprint)
            else:
                logger.debug('fingerprint "{}" had device'.format(fingerprint))

        else:
            logger.debug('request has device "{}"'.format(device))

    else:
        logger.info(
            'request already has device "{}"'.format(device))

    assert device is not None

    set_session_key(request, '__dssoa__device', device)
    set_session_key(request, '__dssoa__device__fingerprint', device.fingerprint)

    return device



import logging
# import inspect
# import warnings

from django.db import transaction
from django.urls import reverse
from django.utils import timezone
from django.utils.translation import activate as activate_translation
from django.core.cache import cache
from django.contrib.auth import (authenticate,
                                 get_backends,
                                 login as django_login)
from django.conf import settings
from django.utils.translation import gettext_lazy as _
from django.core.exceptions import SuspiciousOperation

from rest_framework.renderers import JSONRenderer

from allauth.account.adapter import DefaultAccountAdapter
from allauth.account import app_settings as allauth_app_settings

from .mixins import TryAuthenticateMixin
from .utils import set_session_key, check_user_can_login, invalidate_cookie
from .exceptions import UnsubscribedUserException, DjangoStaffUsersCanNotLoginException  # , ProfileIncompleteException, DectivatedUserException
from .apps.emails.models import EmailAddress
from .apps.devices.models import Device
from .apps.devices.utils import get_request_device
from .apps.services.models import Subscription
from .apps.profiles.utils import get_or_create_user_profile, update_profile
from .apps.passepartout.utils import get_passepartout_login_redirect_url
from .tokens.utils import get_request_jwt_fingerprint, get_request_jwt, renew_response_jwt
from .permissions import is_authenticated, is_staff, is_django_staff
from . import app_settings

logger = logging.getLogger('django_sso_app.core')


class AccountAdapter(DefaultAccountAdapter, TryAuthenticateMixin):
    def __init__(self, request=None):
        super(AccountAdapter, self).__init__(request)

        self.error_messages = DefaultAccountAdapter.error_messages
        self.error_messages['email_not_verified'] = _('We have sent an e-mail to you for verification. Follow the link provided to finalize the signup process. Please contact us if you do not receive it within a few minutes.')

    def populate_username(self, request, user):
        # new user has username == email
        user.username = user.email

    def is_open_for_signup(self, request):
        """
        Checks whether or not the site is open for signups.

        Next to simply returning True/False you can also intervene the
        regular flow by raising an ImmediateHttpResponse
        """
        return True

    def is_ajax(self, request):
        is_ajax = request.path.startswith('/api') or \
                  super(AccountAdapter, self).is_ajax(request) or \
                  request.META.get('CONTENT_TYPE', None) == 'application/json'

        return is_ajax

    def ajax_response(self, request, response, redirect_to=None, form=None, data=None):
        # manually setting response renderer
        setattr(response, 'accepted_renderer', JSONRenderer())
        setattr(response, 'accepted_media_type', 'application/json')
        setattr(response, 'renderer_context', {})

        return super(AccountAdapter, self).ajax_response(request, response, redirect_to, form, data)

    def get_login_redirect_url(self, request):
        """
        Returns the default URL to redirect to after logging in.  Note
        that URLs passed explicitly (e.g. by passing along a `next`
        GET parameter) take precedence over the value returned here.
        """

        if request.path.startswith(reverse('account_signup')) or request.path.startswith(reverse('rest_signup')):
            logger.info('do not get redirect_url from signup')
            redirect_url = None
        else:
            redirect_url = get_passepartout_login_redirect_url(request)

            if redirect_url is None:
                redirect_url = super(AccountAdapter, self).get_login_redirect_url(request)

        return redirect_url

    def login(self, request, user):
        # HACK: This is not nice. The proper Django way is to use an
        # authentication backend
        logger.debug('adapter login')

        # check_user_can_login(user)

        if not hasattr(user, 'backend'):
            from .authentication.backends.login import DjangoSsoAppLoginAuthenticationBackend
            backends = get_backends()
            backend = None
            for b in backends:
                if isinstance(b, DjangoSsoAppLoginAuthenticationBackend):
                    # prefer our own backend
                    backend = b
                    break
                elif not backend and hasattr(b, 'get_user'):
                    # Pick the first vald one
                    backend = b
            backend_path = '.'.join([backend.__module__,
                                     backend.__class__.__name__])
            user.backend = backend_path

        if request.path.startswith(reverse('account_signup')) or request.path.startswith(reverse('rest_signup')):
            logger.debug('do not login from signup')

        else:
            logger.debug('django loggin in')

            # disabling rev update
            setattr(user, '__dssoa__user_loggin_in', True)

            # django
            django_login(request, user)

            if is_authenticated(user):
                if is_django_staff(user):
                   raise DjangoStaffUsersCanNotLoginException()
                else:
                    # setting up request user
                    setattr(request, 'user', user)

    def logout(self, request):
        logger.debug('adapter logout, user logged out, setting flags')
        # print(inspect.stack()[1].function)
        try:
            request_device = get_request_device(request)

        except Exception as e:
            logger.warning('No device on request because of "{}"'.format(e))
            request_device = None

        logging_out_user = request.user

        # allauth logout
        super(AccountAdapter, self).logout(request)

        if not is_authenticated(request.user):
            if app_settings.LOGOUT_DELETES_ALL_PROFILE_DEVICES:
                deleted_devices = self.remove_all_user_profile_devices(logging_out_user)
            else:
                if request_device is not None:
                    deleted_devices = self.remove_user_profile_device(request_device)
                else:
                    deleted_devices = 0

            logger.info('({}) devices deleted for user "{}"'.format(deleted_devices, logging_out_user))

    def confirm_email(self, request, email_address):
        """
        Marks the email address as confirmed on the db,
        deletes old user emails.
        """
        email_address.verified = True
        email_address.set_as_primary(conditional=False)
        email_address.save()

        # one email per user
        email_address.user.emailaddress_set.exclude(email=email_address).delete()

    def set_password(self, user, password):
        user.set_password(password)
        user.save()

    def authenticate(self, request, **credentials):
        logger.debug('adapter authenticate')
        #exceptions print(inspect.stack()[1].function)

        """Only authenticates, does not actually login. See `login`"""
        from .authentication.backends.login import DjangoSsoAppLoginAuthenticationBackend

        self.pre_authenticate(request, **credentials)

        DjangoSsoAppLoginAuthenticationBackend.unstash_authenticated_user()
        user = authenticate(request, **credentials)
        alt_user = DjangoSsoAppLoginAuthenticationBackend.unstash_authenticated_user()
        user = user or alt_user

        if user and allauth_app_settings.LOGIN_ATTEMPTS_LIMIT:
            cache_key = self._get_login_attempts_cache_key(
                request, **credentials)
            cache.delete(cache_key)
        else:
            self.authentication_failed(request, **credentials)

        # check user can login
        authentication_success = False

        try:
            check_user_can_login(user, skip_profile_checks=True)

        except UnsubscribedUserException:
            user_unsubscribed_at = user.sso_app_profile.unsubscribed_at
            if user_unsubscribed_at is not None:
                set_session_key(request, '__dssoa__user_is_unsubscribed', str(user_unsubscribed_at))

        except Exception as e:
            logger.exception('adapter authentication failed "{}"'.format(str(e)))

        else:
            authentication_success = True

        if authentication_success:
            if is_authenticated(user):
                logger.info('adapter authenticated "{}"'.format(user))
                _profile = get_or_create_user_profile(user)

                return user
        else:
            self.authentication_failed(request, **credentials)

    @transaction.atomic
    def save_user(self, request, user, form, commit=True):
        """
        Saves a new User instance using information provided in the
        signup form.
        """
        return super(AccountAdapter, self).save_user(request, user, form, commit)

    def render_mail(self, template_prefix, email, context):
        user = context.get('user')
        logger.info('render mail for "{}"'.format(user))
        user_language = user.sso_app_profile.language

        activate_translation(user_language)
        context.update({
            'EMAILS_DOMAIN': settings.EMAILS_DOMAIN,
            'EMAILS_SITE_NAME': settings.EMAILS_SITE_NAME
        })

        return super(AccountAdapter, self).render_mail(template_prefix, email, context)

    def get_email_confirmation_url(self, request, emailconfirmation):
        """Constructs the email confirmation (activation) url.
        Note that if you have architected your system such that email
        confirmations are sent outside of the request context request
        can be None here.
        """
        _url = reverse(
            "account_confirm_email",
            args=[emailconfirmation.key])

        email_confirmation_url = '{}://{}{}'.format(app_settings.HTTP_PROTOCOL,
                                                    settings.EMAILS_DOMAIN,
                                                    _url)

        return email_confirmation_url

    # pai

    def unconfirm_all_user_emails(self, user):
        logger.info('Unconfirming all emails for user "{}"'.format(user))

        return EmailAddress.objects.filter(user=user,
                                           verified=True).update(verified=False)

    def add_user_profile_device(self, user, fingerprint, secret=None):
        profile = user.sso_app_profile
        secret = secret or app_settings.TOKENS_JWT_SECRET

        logger.info(
            'Adding User Device for profile "{}" with fingerprint "{}" and secret "{}"'.format(profile,
                                                                                               fingerprint,
                                                                                               secret))

        device = Device.objects.create(profile=profile, fingerprint=fingerprint, apigw_jwt_secret=secret)

        logger.debug('device "{}" created with key "{}" secret "{}" and fingerprint "{}"'.format(device,
                                                                                     device.apigw_jwt_key,
                                                                                     device.apigw_jwt_secret,
                                                                                     device.fingerprint))

        return device

    def remove_user_profile_device(self, device):
        logger.info('Removing Device {0}'.format(device.id))

        device.delete()

        return 1

    def remove_all_user_profile_devices(self, user):
        logger.info('Removing All Profile Devices for "{}"'.format(user))

        removed = 0
        for device in user.sso_app_profile.devices.all():
            removed += self.remove_user_profile_device(device)

        return removed

    def subscribe_profile_to_service(self, profile, service, update_rev=False, commit=True):
        logger.info('Subscribinig "{}" to service "{}"'.format(profile, service))

        subscription, _subscription_created = \
            Subscription.objects.get_or_create(
                profile=profile, service=service)

        if _subscription_created:
            logger.info('Created service subscrption for "{}"'.format(profile))

            setattr(profile.user, '__dssoa__subscription_updated', True)

            if update_rev:
                profile.update_rev(commit)
        else:
            logger.info('Profile "{}" can not subscribe "{}"'.format(profile, service))

        return _subscription_created

    def disable_profile_subscription(self, profile, subscription, update_rev=False, commit=True):
        service = subscription.service
        logger.info('Unubscribinig "{}" form service "{}"'.format(profile, service))

        try:
            subscription.unsubscribed_at = timezone.now()
            subscription.save()

            if update_rev and commit:
                profile.save()

        except:
            logger.exception('Profile "{}" can not unsubscribe from "{}"'.format(profile, service))
            return False

        else:
            return True

    def get_request_profile_device(self, request):
        received_jwt = get_request_jwt_fingerprint(request)

        return Device.objects.filter(fingerprint=received_jwt).first()

    def delete_request_profile_device(self, request):
        profile = request.user.sso_app_profile
        logger.info('Deleting request profile device for "{}"'.format(profile))

        removing_device = self.get_request_profile_device(request)

        if removing_device is not None:
            logger.info('Removing logged out user profile device "{}" for "{}"'.format(removing_device, profile))
            self.remove_user_profile_device(removing_device)
        else:
            logger.warning('Can not find request device for profile "{}"'.format(profile))

    def try_completely_unsubscribe(self, username, email, password):
        if self.request is None:
            raise KeyError('no request object')

        user = self.try_authenticate(username, email, password)
        requesting_user = self.request.user

        if not is_django_staff(user):
            profile = user.sso_app_profile
            logger.info('Completely unsubscribing Profile "{}"'.format(profile))

            profile.unsubscribed_at = timezone.now()
            profile.is_active = False
            user.is_active = False

            user.save()  # Profile update managed by user signal

            self.logout(self.request)
        else:
            _msg = 'User "{}" tried to completely unsubscribe "{}"'.format(requesting_user, user)
            logger.warning(_msg)

            raise SuspiciousOperation(_msg)

    def update_user_email(self, user, new_email):
        user.email = new_email  # set user model email

        setattr(user, '__dssoa__email_updated', True)

        user.save()

    def update_user_username(self, user, new_username):
        user.username = new_username # set user model username

        setattr(user, '__dssoa__username_updated', True)

        user.save()

    def try_disable_response_jwt(self, user, response):
        if self.request is None:
            raise KeyError('no request object')

        invalidate_cookie(response, app_settings.JWT_COOKIE_NAME)

    def try_update_response_jwt(self, user, response):
        if self.request is None:
            raise KeyError('no request object')

        logger.info('Updating resopnse jwt for "{}"'.format(user))

        requesting_user = self.request.user

        if not is_django_staff(requesting_user):
            received_jwt = get_request_jwt(self.request)

            if received_jwt is not None:
                is_same_user = requesting_user == user

                if is_same_user:
                    logger.info('From browser as same user "{}"'.format(user))
                    renew_response_jwt(received_jwt, user, self.request, response)

                elif not is_same_user and is_staff(requesting_user):
                    # returns user jwt to staff user
                    logger.info('From browser as staff user "{}"'.format(requesting_user))
                    renew_response_jwt(received_jwt, user, self.request, response)

                else:
                    _msg = 'User "{}" tried to updated profile for "{}"'.format(requesting_user, user)
                    logger.warning(_msg)

                    raise SuspiciousOperation(_msg)

        else:
            _msg = 'Admin user "{}" updated profile for "{}", skipping jwt update'.format(requesting_user, user)
            logger.info(_msg)



import logging

from rest_framework.reverse import reverse
from rest_framework import serializers

from ...serializers import AbsoluteUrlSerializer
from .models import Service, Subscription

logger = logging.getLogger('django_sso_app.core.apps.services')


class ServiceSerializer(AbsoluteUrlSerializer):
    is_subscribed = serializers.SerializerMethodField(required=False)
    is_unsubscribed = serializers.SerializerMethodField(required=False)
    tos = serializers.SerializerMethodField(required=False)
    subscribe = serializers.SerializerMethodField(required=False)
    unsubscribe = serializers.SerializerMethodField(required=False)

    class Meta:
        model = Service
        read_only_fields = ('url', 'id', 'picture', 'is_subscribed', 'is_unsubscribed')
        fields = read_only_fields + ('service_url', 'name',
                                     'cookie_domain',
                                     'redirect_wait',
                                     'is_public',
                                     'tos',
                                     'subscription_required',
                                     'subscribe', 'unsubscribe')  # 'groups'

    def get_is_subscribed(self, instance):
        request = self.context.get("request")
        user = request.user
        if user is None or user.is_anonymous:
            return False

        subscription = request.user.sso_app_profile.subscriptions.filter(service=instance).first()
        if subscription is not None:
            return subscription.is_subscribed

        return False

    def get_is_unsubscribed(self, instance):
        request = self.context.get("request")
        user = request.user

        if user is None or user.is_anonymous:
            return False

        subscription = request.user.sso_app_profile.subscriptions.filter(service=instance).first()
        if subscription is not None:
            return subscription.is_unsubscribed

        return False

    def get_tos(self, instance):
        request = self.context.get('request')
        language = request.LANGUAGE_CODE
        tos = instance.get_tos(language)

        if tos is None:
            instance.get_tos('en')
        if tos is None:
            return None

        return tos.text

    def get_subscribe(self, instance):
        request = self.context['request']

        return request.build_absolute_uri(reverse('django_sso_app_service:rest-subscription',
                                          kwargs={'pk': instance.id}))

    def get_unsubscribe(self, instance):
        request = self.context['request']

        return request.build_absolute_uri(reverse('django_sso_app_service:rest-unsubscription',
                                          kwargs={'pk': instance.id}))


class ServiceSubscriptionSerializer(serializers.Serializer):
    sso_id = serializers.CharField(required=False)


class SubscriptionSerializer(AbsoluteUrlSerializer):
    service = ServiceSerializer()
    profile = serializers.SerializerMethodField(required=False)

    class Meta:
        model = Subscription
        fields = ('url', 'id', 'service', 'created_at', 'is_unsubscribed', 'profile', 'is_public')

    def get_profile(self, instance):
        request = self.context['request']
        profile_url = instance.profile.get_absolute_rest_url()

        return request.build_absolute_uri(profile_url)



import logging

from django.contrib.auth.models import Group as DjangoGroup
from django.utils.translation import ugettext_lazy as _
from django.urls import reverse

logger = logging.getLogger('django_sso_app.core.apps.groups')


class Group(DjangoGroup):
    class Meta:
        proxy = True
        app_label = 'django_sso_app'
        verbose_name = _('Group')

    def get_absolute_rest_url(self):
        return reverse("django_sso_app_group:rest-detail", args=[self.pk])



from django.contrib.auth import get_user_model

from django_sso_app.core.tests.factories import UserTestCase


class ProfileTestCase(UserTestCase):

    def test_profile_field_update_updates_rev(self):
        """
        User update updates user revision
        """
        user = self._get_new_user()
        profile = user.sso_app_profile

        profile_rev = profile.sso_rev

        profile.country = 'RU'
        profile.save()

        profile.refresh_from_db()

        self.assertEqual(profile.sso_rev, profile_rev + 1)

    def test_django_user_email_update_updates_profile_django_user_email(self):
        User = get_user_model()

        user = self._get_new_user()
        profile = user.sso_app_profile

        profile_rev = profile.sso_rev

        self.assertEqual(user.email, profile.django_user_email)

        user = User.objects.get(username=user.username)
        user.email = self._get_random_email()
        user.save()

        profile.refresh_from_db()

        self.assertEqual(profile.django_user_email, user.email)
        self.assertEqual(profile.sso_rev, profile_rev + 1)

    def test_django_user_username_update_updates_profile_django_user_username(self):
        User = get_user_model()

        user = self._get_new_user()
        profile = user.sso_app_profile

        profile_rev = profile.sso_rev

        self.assertEqual(user.username, profile.django_user_username)

        user = User.objects.get(username=user.username)
        user.username = self._get_random_username()
        user.save()

        profile.refresh_from_db()

        self.assertEqual(profile.django_user_username, user.username)
        self.assertEqual(profile.sso_rev, profile_rev + 1)



import logging
import requests

from django.contrib.auth import get_user_model

from .... import app_settings
from .functions import get_apigateway_consumer_id

logger = logging.getLogger('django_sso_app.core.apps.api_gateway.kong')
User = get_user_model()


def get_apigw_consumer(profile):
    logger.info('getting apigw consumer for {}'.format(profile))

    consumer_id = get_apigateway_consumer_id(profile.sso_id)

    url = app_settings.APIGATEWAY_HOST + "/consumers/{}".format(consumer_id)
    r = requests.get(url)

    response_body = None
    status_code = r.status_code
    try:
        response_body = r.json()
    except:
        logger.info('Response ({}) has no payload'.format(status_code))

    logger.info('apigw response ({0}) {1}'.format(status_code, response_body))
    return status_code, response_body


def create_apigw_consumer(profile):
    logger.info('creating apigw consumer for {}'.format(profile))

    consumer_id = get_apigateway_consumer_id(profile.sso_id)

    url = app_settings.APIGATEWAY_HOST + "/consumers/"
    data = {'username': consumer_id}

    r = requests.post(url, json=data)

    response_body = None
    status_code = r.status_code
    try:
        response_body = r.json()
    except:
        logger.info('Response ({}) has no payload'.format(status_code))

    logger.info('apigw response ({0}) {1}'.format(status_code, response_body))
    return status_code, response_body


def delete_apigw_consumer(profile):
    consumer_id = get_apigateway_consumer_id(profile.sso_id)
    logger.info('deleting apigw consumer for {}'.format(profile))

    url = app_settings.APIGATEWAY_HOST + "/consumers/" + consumer_id

    r = requests.delete(url)

    response_body = None
    status_code = r.status_code
    try:
        response_body = r.json()
    except:
        logger.info('Response ({}) has no payload'.format(status_code))

    logger.info('apigw response ({0}) {1}'.format(status_code, response_body))
    return status_code, response_body


def create_apigw_consumer_jwt(profile):
    consumer_id = get_apigateway_consumer_id(profile.sso_id)
    logger.info('creating apigw consumer jwt for {}'.format(profile))

    url = app_settings.APIGATEWAY_HOST + "/consumers/" + consumer_id + "/jwt/"
    data = {}

    r = requests.post(url, json=data)

    response_body = None
    status_code = r.status_code
    try:
        response_body = r.json()
    except:
        logger.info('Response ({}) has no payload'.format(status_code))

    logger.info('apigw response ({0}) {1}'.format(status_code, response_body))
    return status_code, response_body


def delete_apigw_consumer_jwt(profile, jwt_id):
    consumer_id = get_apigateway_consumer_id(profile.sso_id)
    logger.info('deleting apigw consumer jwt for {}'.format(profile))

    url = '{}/{}/{}/jwt/{}'.format(app_settings.APIGATEWAY_HOST, "consumers", consumer_id, jwt_id)

    logger.info('calling kong url "{}"'.format(url))

    r = requests.delete(url)

    status_code = r.status_code

    logger.info('apigw response ({0}) {1}'.format(status_code, None))
    return r.status_code, None


def get_apigw_consumer_jwts(profile):
    consumer_id = get_apigateway_consumer_id(profile.sso_id)
    logger.info('getting apigw consumer jwts for {}'.format(profile))

    url = app_settings.APIGATEWAY_HOST + "/consumers/" + consumer_id + "/jwt/"

    r = requests.get(url)

    response_body = None
    status_code = r.status_code
    try:
        response_body = r.json()
    except:
        logger.info('Response ({}) has no payload'.format(status_code))

    logger.info('apigw response ({0}) {1}'.format(status_code, response_body))
    return status_code, response_body


def create_apigw_consumer_acl(profile, group_name=app_settings.APIGATEWAY_CONSUMER_GROUP):
    consumer_id = get_apigateway_consumer_id(profile.sso_id)
    logger.info('creating apigw consumer acl {} for {}'.format(group_name,
                                                               profile))

    url = app_settings.APIGATEWAY_HOST + "/consumers/" + consumer_id + "/acls/"
    data = {"group": group_name}

    r = requests.post(url, json=data)

    response_body = None
    status_code = r.status_code
    try:
        response_body = r.json()
    except:
        logger.info('Response ({}) has no payload'.format(status_code))

    logger.info('apigw response ({0}) {1}'.format(status_code, response_body))
    return status_code, response_body


def get_apigw_consumer_acls(profile):
    consumer_id = get_apigateway_consumer_id(profile.sso_id)
    logger.info('getting apigw consumer groups for {}'.format(profile))

    url = app_settings.APIGATEWAY_HOST + "/consumers/" + consumer_id + "/acls/"

    r = requests.get(url)

    response_body = None
    status_code = r.status_code
    try:
        response_body = r.json()
    except:
        logger.info('Response ({}) has no payload'.format(status_code))

    logger.info('apigw response ({0}) {1}'.format(status_code, response_body))
    return status_code, response_body


def delete_apigw_consumer_acl(profile, group_name):
    consumer_id = get_apigateway_consumer_id(profile.sso_id)
    logger.info('deleting apigw consumer acl {} for {}'.format(group_name,
                                                               profile))

    url = app_settings.APIGATEWAY_HOST + "/consumers/" + consumer_id + "/acls/" + group_name

    r = requests.delete(url)

    status_code = r.status_code

    logger.info('apigw response ({0}) {1}'.format(status_code, None))
    return r.status_code, None


def get_profile_apigw_consumer_id(profile, force_recreate=False):

    if profile.apigw_consumer_id is None or force_recreate:
        logger.info('Profile "{}" has no apigateway_consumer_id, checking if exists or creating new'.format(profile))

        # try to find an existent consumer
        status_code_0, consumer = get_apigw_consumer(profile)
        if status_code_0 == 404:
            logger.info('No apigw consumer found for profile "{}", creating'.format(profile))

            # creating new
            status_code_1, consumer = create_apigw_consumer(profile)
            if status_code_1 != 201:
                logger.error('Error ({0}) creating apigw consumer, {1}'.format(
                    status_code_1, consumer))
                raise Exception(
                    "Error ({0}) creating apigw consumer, {1}".format(
                        status_code_1, consumer))

        elif status_code_0 == 200:
            logger.info('Consumer "{}" found for profile "{}"'.format(consumer, profile))
        else:
            raise Exception(
                "Error ({0}) getting apigw consumer, {1}".format(status_code_0, consumer))

        # adding profile to groups
        profile_groups = [app_settings.APIGATEWAY_CONSUMER_GROUP] + \
                         list(profile.user.groups.values_list('name', flat=True))

        for group_name in profile_groups:
            status_code_2, acl = create_apigw_consumer_acl(profile, group_name)
            if status_code_2 != 201:
                # delete_apigw_consumer(username)
                logger.error(
                    'Error ({0}) creating apigw consumer acl, {1}'.format(
                        status_code_2, acl))
                raise Exception(
                    "Error {0} creating apigw consumer acl, {1}".format(
                        status_code_2, acl))

        # Do not update sso_rev
        setattr(profile.user, '__dssoa__apigateway_update', True)

        # saving profile
        profile.apigw_consumer_id = consumer['id']
        profile.save()

    else:
        logger.info('Profile {} already has apigateway_consumer_id'.format(profile))

    return profile.apigw_consumer_id



from ...serializers import AbsoluteUrlSerializer
from .models import Group


class GroupSerializer(AbsoluteUrlSerializer):
    class Meta:
        model = Group
        read_only_fields = ('url', 'id', 'name')
        fields = read_only_fields



from django.conf.urls import url
from rest_framework.urlpatterns import format_suffix_patterns

from .views import ServiceApiViewSet, SubscriptionApiViewSet, ServiceSubscriptionApiViewSet


_service__urlpatterns = [
    url(r'^$',
        ServiceApiViewSet.as_view({'get': 'list'}),
        name='rest-list'),
    url(r'^(?P<pk>\d+)/$',
        ServiceApiViewSet.as_view({'get': 'retrieve'}),
        name='rest-detail'),

    url(r'^(?P<pk>\d+)/subscribe/$',
        ServiceSubscriptionApiViewSet.as_view({'post': 'subscribe'}),
        name='rest-subscription'),

    url(r'^(?P<pk>\d+)/unsubscribe/$',
        ServiceSubscriptionApiViewSet.as_view({'post': 'unsubscribe'}),
        name='rest-unsubscription'),
]

_subscription_urlpatterns = [
    url(r'^subscriptions/$',
        SubscriptionApiViewSet.as_view({'get': 'list'}),
        name='rest-subscription-list'),

    url(r'^subscriptions/(?P<pk>\d+)/$',
        SubscriptionApiViewSet.as_view({'get': 'retrieve'}),
        name='rest-subscription-detail'),
]

service_urlpatterns = format_suffix_patterns(_service__urlpatterns)
subscription_urlpatterns = format_suffix_patterns(_subscription_urlpatterns)

urlpatterns = (service_urlpatterns + subscription_urlpatterns, 'django_sso_app_service')



import json

from django.contrib.auth import get_user_model

from django.urls import reverse

from rest_framework import status
from django_sso_app.core.tests.factories import UserTestCase, responses

from django_sso_app.core.apps.profiles.models import Profile


class TestStatus(UserTestCase):

    def setUp(self):
        self.new_pass = self._get_random_pass()
        self.new_user = self._get_new_user(password=self.new_pass)
        self.new_user_profile = self.new_user.sso_app_profile
        self.new_user_profile.groups.add(self._get_new_group())

    @responses.activate
    def test_switch_between_statuses(self):
        with self.settings(DJANGO_SSO_APP_SHAPE='backend_only'):

            response = self._get_client().post(
                reverse('rest_login'),
                data=json.dumps(self._get_login_object(self.new_user.email, self.new_pass)),
                content_type='application/json'
            )

            self.assertEqual(response.status_code, status.HTTP_200_OK, 'new user can not login')

        with self.settings(DJANGO_SSO_APP_SHAPE='backend_app'):
            response2 = self._get_client().post(
                reverse('rest_login'),
                data=json.dumps(self._get_login_object(self.new_user.email, self.new_pass)),
                content_type='application/json'
            )

            self.assertEqual(response2.status_code, status.HTTP_200_OK, 'new user can not login')

        with self.settings(DJANGO_SSO_APP_SHAPE='app',
                           DJANGO_SSO_APP_SERVICE_URL='http://example.com'):

            device = self.new_user_profile.devices.first()

            jwt_cookie = self._get_jwt_cookie(device)

            client = self._get_client()
            client.cookies = jwt_cookie

            response = client.get(
                reverse('django_sso_app_profile:rest-detail', args=(self.new_user_profile.sso_id,)),
                content_type='application/json'
            )

            self.assertEqual(response.status_code, status.HTTP_200_OK, 'unable to get profile detail')

            self.assertEqual(Profile.objects.filter(sso_id=self.new_user_profile.sso_id).count(), 1,
                             'no profile model created')

            self.assertEqual(response.data.get('sso_id'), str(self.new_user_profile.sso_id), 'sso_id differs')

        with self.settings(DJANGO_SSO_APP_SHAPE='app_persistence',
                           DJANGO_SSO_APP_SERVICE_URL='http://example.com'):

            device = self.new_user_profile.devices.first()
            jwt_cookie = self._get_jwt_cookie(device)

            client = self._get_client()
            client.cookies = jwt_cookie

            response = client.get(
                reverse('django_sso_app_profile:rest-detail', args=(self.new_user_profile.sso_id,)),
                content_type='application/json'
            )

            self.assertEqual(response.status_code, status.HTTP_200_OK, 'unable to get profile detail')

            self.assertEqual(Profile.objects.filter(sso_id=self.new_user_profile.sso_id).count(), 1,
                             'no profile model created')

            self.assertEqual(response.data.get('sso_id'), str(self.new_user_profile.sso_id), 'sso_id differs')

        with self.settings(DJANGO_SSO_APP_SHAPE='app_apigateway',
                           DJANGO_SSO_APP_SERVICE_URL='http://example.com'):

            device = self.new_user_profile.devices.first()
            jwt_cookie = self._get_jwt_cookie(device)

            remote_profile_uuid = self.new_user_profile.sso_id
            remote_user_group_name = self.new_user_profile.groups.first().name

            profile_url = reverse('django_sso_app_profile:rest-detail', args=(remote_profile_uuid,))

            client = self._get_client()
            client.cookies = jwt_cookie

            response = client.get(
                profile_url,
                content_type='application/json',
                HTTP_X_CONSUMER_USERNAME=remote_profile_uuid.replace('-', '_'),
                HTTP_X_CONSUMER_GROUPS=remote_user_group_name
            )

            self.assertEqual(response.status_code, status.HTTP_200_OK)
            self.assertEqual(response.data.get('sso_id'), remote_profile_uuid, 'sso_id differs from header')

        with self.settings(DJANGO_SSO_APP_SHAPE='app_persistence_apigateway',
                           DJANGO_SSO_APP_SERVICE_URL='http://example.com'):

            device = self.new_user_profile.devices.first()
            jwt_cookie = self._get_jwt_cookie(device)

            remote_profile_uuid = self.new_user_profile.sso_id
            remote_user_group_name = self.new_user_profile.groups.first().name

            profile_url = reverse('django_sso_app_profile:rest-detail', args=(remote_profile_uuid,))

            client = self._get_client()
            client.cookies = jwt_cookie

            response = client.get(
                profile_url,
                content_type='application/json',
                HTTP_X_CONSUMER_USERNAME=remote_profile_uuid.replace('-', '_'),
                HTTP_X_CONSUMER_GROUPS=remote_user_group_name
            )

            self.assertEqual(response.status_code, status.HTTP_200_OK, 'unable to get profile detail')

            self.assertEqual(Profile.objects.filter(sso_id=self.new_user_profile.sso_id).count(), 1,
                             'no profile model created')

            self.assertEqual(response.data.get('sso_id'), str(self.new_user_profile.sso_id), 'sso_id differs')



from allauth.exceptions import ImmediateHttpResponse


class BaseException(Exception):
    def __init__(self, msg=None):
        self.msg = str(msg)


class DjangoStaffUsersCanNotLoginException(BaseException):
    pass


class AuthenticationFailed(BaseException):
    pass


class InvalidToken(BaseException):
    pass


class TokenError(BaseException):
    pass


class ProfileIncompleteException(ImmediateHttpResponse):
    pass


class DectivatedUserException(Exception):
    pass


class UnsubscribedUserException(Exception):
    pass


class ServiceSubscriptionRequiredException(ImmediateHttpResponse):
    pass


class RequestHasValidJwtWithNoDeviceAssociated(BaseException):
    pass


class UndefinedRequestUserException(BaseException):
    pass



import logging

# import pyximport
# pyximport.install()

from ...permissions import is_django_staff
from ... import app_settings
from .models import Profile as Profile

logger = logging.getLogger('django_sso_app.core.apps.profiles')


def update_profile(profile, update_object, commit=True):
    logger.info('Updating profile fields for "{}"'.format(profile))

    for f in app_settings.PROFILE_FIELDS + ('sso_rev', 'meta'):
        if hasattr(profile, f):
            new_val = update_object.get(f, None)
            if new_val is not None:
                setattr(profile, f, new_val)
                logger.debug('profile field updated "{}":"{}"'.format(f, new_val if f != 'password' else '*'))

    if commit:
        profile.save()

    return profile


def get_or_create_user_profile(user, initials=None, commit=True):
    logger.debug('get or create user profile')

    try:
        profile = user.sso_app_profile

    except Profile.DoesNotExist:
        logger.info('User "{}" has no profile, creating one'.format(user))

        profile = Profile(user=user, django_user_email=user.email, django_user_username=user.username)

        sso_id = getattr(user, '__dssoa__profile__sso_id', None)
        if sso_id is not None:
            sso_rev = getattr(user, '__dssoa__profile__sso_rev', 0)
            profile.sso_id = sso_id
            profile.sso_rev = sso_rev
        else:
            # django staff users sso_id equals username
            if is_django_staff(user):
                profile.sso_id = user.username
                profile.sso_rev = 0

                setattr(user, '__dssoa__creating', True)

                profile.save()

        profile_initials = initials or getattr(user, '__dssoa__profile__object', None)

        if profile_initials is not None:
            logger.debug('Got profile initials "{}"'.format(profile_initials))
            profile = update_profile(profile, profile_initials, commit=False)
        else:
            logger.debug('No profile initials')

        if commit:
            profile.save()

    # force user sso_app_profile object field
    setattr(user, 'sso_app_profile', profile)

    return profile



from django.apps import AppConfig


class DjangoSsoAppConfig(AppConfig):
    name = 'django_sso_app'
    label = 'django_sso_app'

    def ready(self):
        try:
            import django_sso_app.core.apps.api_gateway.signals  # noqa F401
            import django_sso_app.core.apps.emails.signals  # noqa F401
            import django_sso_app.core.apps.devices.signals  # noqa F401
            import django_sso_app.core.apps.groups.signals  # noqa F401
            import django_sso_app.core.apps.passepartout.signals  # noqa F401
            import django_sso_app.core.apps.profiles.signals  # noqa F401
            import django_sso_app.core.apps.services.signals  # noqa F401
            import django_sso_app.core.apps.status.signals  # noqa F401
            import django_sso_app.core.apps.users.signals  # noqa F401

        except ImportError:
            pass



from .settings.common import *


class AppSettings(object):

    def __init__(self, prefix):
        self.prefix = prefix
        self.env = env  # got from .settings.common
        # check shape string
        if self.SHAPE not in AVAILABLE_SHAPES:
            raise Exception('Wrong DJANGO_SSO_APP_SHAPE ({})'.format(self.SHAPE))

    def _env(self, key, dflt, type=str):
        if type == bool:
            return self.env.bool(key, default=dflt)
        elif type == tuple:
            return self.env.tuple(key, default=dflt)
        if type == int:
            return self.env.int(key, default=dflt)
        elif type == list:
            return self.env.list(key, default=dflt)
        else:
            return self.env(key, default=dflt)

    def _django_setting(self, name, dflt, type=str):
        from django.conf import settings
        getter = getattr(settings,
                         'DJANGO_SETTING_GETTER',
                         lambda name, dflt: getattr(settings, name, self._env(name, dflt, type)))
        return getter(name, dflt)

    def _allauth_setting(self, name, dflt):
        from django.conf import settings
        getter = getattr(settings,
                         'ALLAUTH_SETTING_GETTER',
                         lambda name, dflt: getattr(settings, name, dflt))
        return getter(self.prefix + name, dflt)

    def _setting(self, name, dflt, type=str):
        from django.conf import settings
        getter = getattr(settings,
                         'DJANGO_SSO_APP_SETTING_GETTER',
                         lambda name, dflt: getattr(settings, name, self._env(name, dflt, type)))
        return getter(self.prefix + name, dflt)

    @property
    def SHAPE(self):
        return self._setting('SHAPE', DJANGO_SSO_APP_SHAPE)

    @property
    def BACKEND_ENABLED(self):
        return self.SHAPE in BACKEND_SHAPES

    @property
    def APP_ENABLED(self):
        return not self.BACKEND_ENABLED

    @property
    def APIGATEWAY_ENABLED(self):
        return self.SHAPE.find('apigateway') > -1

    @property
    def APIGATEWAY_HOST(self):
        return self._setting('APIGATEWAY_HOST', 'http://kong:8001')

    @property
    def BACKEND_STANDALONE(self):
        return self.SHAPE.find('standalone') > -1

    @property
    def REPLICATE_PROFILE(self):
        return self.SHAPE.find('persistence') > -1

    @property
    def BACKEND_CUSTOM_FRONTEND_APP(self):
        return self._setting('BACKEND_CUSTOM_FRONTEND_APP', None)

    @property
    def BACKEND_HAS_CUSTOM_FRONTEND_APP(self):
        return self.BACKEND_CUSTOM_FRONTEND_APP is not None

    @property
    def BACKEND_FRONTEND_APP_RESTONLY(self):
        return self._setting('BACKEND_FRONTEND_APP_RESTONLY', self.BACKEND_HAS_CUSTOM_FRONTEND_APP)

    @property
    def COOKIE_DOMAIN(self):
        return self._django_setting('COOKIE_DOMAIN', COOKIE_DOMAIN)  # defaults to .settings value

    @property
    def COOKIE_AGE(self):
        days_expire = 365
        return self._django_setting('COOKIE_AGE', days_expire * 24 * 60 * 60, int)

    @property
    def COOKIE_HTTPONLY(self):
        return self._django_setting('COOKIE_HTTPONLY', True, bool)

    @property
    def APIGATEWAY_CONSUMER_GROUP(self):
        return self._setting('APIGATEWAY_CONSUMER_GROUP', self.COOKIE_DOMAIN)

    @property
    def APIGATEWAY_CONSUMER_USERNAME_HEADER(self):
        return self._setting('APIGATEWAY_CONSUMER_USERNAME_HEADER', 'HTTP_X_CONSUMER_USERNAME')

    @property
    def APIGATEWAY_ANONYMOUS_CONSUMER_USERNAME(self):
        return self._setting('APIGATEWAY_ANONYMOUS_CONSUMER_USERNAME', 'anonymous')

    @property
    def APIGATEWAY_ANONYMOUS_CONSUMER_USERNAMES(self):
        return self._setting('APIGATEWAY_ANONYMOUS_CONSUMER_USERNAMES', [self.APIGATEWAY_ANONYMOUS_CONSUMER_USERNAME])

    @property
    def APIGATEWAY_CONSUMER_GROUPS_HEADER(self):
        return self._setting('APIGATEWAY_CONSUMER_GROUPS_HEADER', 'HTTP_X_CONSUMER_GROUPS')

    @property
    def BACKEND_SIGNUP_MUST_FILL_PROFILE(self):
        return self._setting('BACKEND_SIGNUP_MUST_FILL_PROFILE', False, bool)

    @property
    def USER_FIELDS(self):
        return self._setting('USER_FIELDS', ('username', 'email'), tuple)

    @property
    def REQUIRED_USER_FIELDS(self):
        return self._setting('REQUIRED_USER_FIELDS', self.USER_FIELDS)

    @property
    def PROFILE_FIELDS(self):
        return self._setting('PROFILE_FIELDS', ('first_name', 'last_name',
                                                'ssn', 'phone',
                                                'description',
                                                'picture',
                                                'birthdate',
                                                'latitude', 'longitude',
                                                'country',
                                                'address',
                                                'language',
                                                'alias'), tuple)

    @property
    def REQUIRED_PROFILE_FIELDS(self):
        return self._setting('REQUIRED_PROFILE_FIELDS', ('first_name', ), tuple)

    @property
    def HTTP_PROTOCOL(self):
        return self._django_setting('HTTP_PROTOCOL', ACCOUNT_DEFAULT_HTTP_PROTOCOL)  # defaults to .settings value

    @property
    def APP_DOMAIN(self):
        return self._django_setting('APP_DOMAIN', APP_DOMAIN)  # defaults to .settings value

    @property
    def APP_URL(self):
        return self._django_setting('APP_URL', self.HTTP_PROTOCOL + '://' + self.APP_DOMAIN)

    @property
    def SERVICE_URL(self):
        return self._setting('SERVICE_URL', self.APP_URL)

    @property
    def SERVICE_SUBSCRIPTION_REQUIRED(self):
        return self._setting('SERVICE_SUBSCRIPTION_REQUIRED', self.REPLICATE_PROFILE, bool)

    @property
    def LOGIN_URL(self):
        return self._setting('LOGIN_URL', '/login/')

    @property
    def LOGIN_MUST_FILL_PROFILE(self):
        return self._setting('LOGIN_MUST_FILL_PROFILE', True, bool)

    @property
    def MANAGE_USER_GROUPS(self):
        return self._setting('MANAGE_USER_GROUPS', True, bool)

    @property
    def AUTH_HEADER_TYPES(self):
        return self._setting('AUTH_HEADER_TYPES', ('Bearer',))

    @property
    def JWT_COOKIE_NAME(self):
        return self._setting('JWT_COOKIE_NAME', 'jwt')

    @property
    def JWT_ALGORITHM(self):
        _JWT_ALLOWED_ALGORITHMS = (
            'HS256',
            'HS384',
            'HS512',
            'RS256',
            'RS384',
            'RS512',
        )

        return self._setting('JWT_ALGORITHM', _JWT_ALLOWED_ALGORITHMS[0])

    @property
    def TOKENS_JWT_SECRET(self):
        return self._setting('TOKENS_JWT_SECRET', 'secret')

    @property
    def USER_ID_CLAIM(self):
        return self._setting('USER_ID_CLAIM', 'sso_id')

    @property
    def BACKEND_STAFF_TOKEN(self):
        return self._setting('BACKEND_STAFF_TOKEN', None)

    @property
    def DJANGO_SSO_APP_USER_ID_FIELD(self):
        return self._setting('DJANGO_SSO_APP_USER_ID_FIELD', 'sso_app_profile__sso_id')

    @property
    def BACKEND_HIDE_PASSWORD_FROM_USER_SERIALIZER(self):
        return self._setting('BACKEND_HIDE_PASSWORD_FROM_USER_SERIALIZER', self.BACKEND_ENABLED, bool)

    @property
    def BACKEND_DOMAINS(self):
        return self._setting('BACKEND_DOMAINS', [self.APP_DOMAIN], list)

    @property
    def BACKEND_DOMAIN(self):
        return self._setting('BACKEND_DOMAIN', self.BACKEND_DOMAINS[0],)

    @property
    def BACKEND_DOMAINS_DICT(self):
        _BACKEND_DOMAINS_DICT = {}
        i = 0
        for el in self.BACKEND_DOMAINS:
            _BACKEND_DOMAINS_DICT[el] = i
            i += 1
        return _BACKEND_DOMAINS_DICT

    @property
    def BACKEND_FULL_URLS_CHAIN(self):
        return ['{}://{}'.format(self.HTTP_PROTOCOL, el) for el in self.BACKEND_DOMAINS]

    @property
    def BACKEND_URLS_CHAIN(self):
        return ['{}://{}'.format(self.HTTP_PROTOCOL, el) for el in self.BACKEND_DOMAINS if el != self.APP_DOMAIN]

    @property
    def BACKEND_URLS_CHAIN_DICT(self):
        i = 0
        _BACKEND_URLS_CHAIN_DICT = {}
        for el in self.BACKEND_URLS_CHAIN:
            _BACKEND_URLS_CHAIN_DICT[el] = i
            i += 1
        return _BACKEND_URLS_CHAIN_DICT

    @property
    def PASSEPARTOUT_PROCESS_ENABLED(self):
        return len(self.BACKEND_URLS_CHAIN) > 0

    @property
    def BACKEND_URL(self):
        return self._setting('BACKEND_URL', self.HTTP_PROTOCOL + '://' + self.BACKEND_DOMAIN)

    @property
    def PROFILE_UPDATE_URL(self):
        return self._setting('PROFILE_UPDATE_URL', '/profile/update/')

    @property
    def PROFILE_COMPLETE_URL(self):
        return self._setting('PROFILE_COMPLETE_URL', '/profile/complete/')

    @property
    def REMOTE_USERS_URL(self):
        return self.BACKEND_URL + '/api/v1/auth/users/'

    @property
    def REMOTE_USER_URL(self):
        return self.REMOTE_USERS_URL + '{sso_id}/'

    @property
    def REMOTE_PROFILES_URL(self):
        return self.BACKEND_URL + '/api/v1/auth/profiles/'

    @property
    def REMOTE_PROFILE_URL(self):
        return self.REMOTE_PROFILES_URL + '{sso_id}/'

    @property
    def REMOTE_LOGIN_URL(self):
        return self.BACKEND_URL + '/login/' + '?next=' + self.SERVICE_URL

    @property
    def REMOTE_LOGOUT_URL(self):
        return self.BACKEND_URL + '/logout/' + '?next=' + self.SERVICE_URL

    @property
    def REMOTE_SIGNUP_URL(self):
        return self.BACKEND_URL + '/signup/' + '?next=' + self.SERVICE_URL

    @property
    def REMOTE_EMAIL_URL(self):
        return self.BACKEND_URL + '/email/'

    @property
    def REMOTE_PROFILE_VIEW_URL(self):
        return self.BACKEND_URL + '/profile/'

    @property
    def REMOTE_PROFILE_UPDATE_URL(self):
        return self.BACKEND_URL + '/profile/update/'

    @property
    def REMOTE_PROFILE_COMPLETE_URL(self):
        return self.BACKEND_URL + '/profile/complete/'

    @property
    def I18N_PATH_ENABLED(self):
        return self._django_setting('I18N_PATH_ENABLED', I18N_PATH_ENABLED, bool)

    @property
    def SOCIALACCOUNTS(self):
        return self._django_setting('SOCIALACCOUNTS', ['google'], list)

    @property
    def SOCIALACCOUNTS_ENABLED(self):
        return self._django_setting('SOCIALACCOUNTS_ENABLED', len(self.SOCIALACCOUNTS) > 0, bool)

    @property
    def STAFF_USER_GROUPS(self):
        return self._setting('STAFF_USER_GROUPS', ['staff'], list)

    @property
    def LOGOUT_DELETES_ALL_PROFILE_DEVICES(self):
        return self._setting('LOGOUT_DELETES_ALL_PROFILE_DEVICES', False, bool)

    @property
    def USER_TYPES(self):
        return self._setting('USER_TYPES', [], list)

    @property
    def DEFAULT_USER_GROUPS(self):
        return self._setting('DEFAULT_USER_GROUPS', [], list)

    @property
    def SAME_SITE_COOKIE_NONE(self):
        return self._setting('SAME_SITE_COOKIE_NONE', False, bool)

    @property
    def SIGNALS_DISABLED_COMMANDS(self):
        return self._setting('SIGNALS_DISABLED_COMMANDS', ['loaddata', 'dpb_couchdb_loaddata'], list)

# Ugly? Pennersr says Guido recommends this himself ...
# http://mail.python.org/pipermail/python-ideas/2012-May/014969.html
import sys  # noqa

app_settings = AppSettings('DJANGO_SSO_APP_')
app_settings.__name__ = __name__
sys.modules[__name__] = app_settings



from rest_framework import viewsets
from rest_framework import permissions

from .serializers import GroupSerializer
from .models import Group


class GroupApiViewSet(viewsets.ReadOnlyModelViewSet):

    serializer_class = GroupSerializer
    queryset = Group.objects.all()
    lookup_field = 'pk'
    permission_classes = (permissions.IsAuthenticated, )

    def list(self, request, *args, **kwargs):
        """
        List groups
        """
        return super(GroupApiViewSet, self).list(request, *args, **kwargs)

    def retrieve(self, request, pk=None, *args, **kwargs):
        """
        Group detail
        """
        return super(GroupApiViewSet, self).retrieve(request, pk, *args, **kwargs)



from django.contrib.auth import get_user_model

from ...config import celery_app

User = get_user_model()


@celery_app.task()
def get_users_count():
    """A pointless Celery task to demonstrate usage."""
    return User.objects.count()



import logging

from django.dispatch import receiver
from django.db.models.signals import post_save

from allauth.account.models import EmailAddress

logger = logging.getLogger('django_sso_app.core.apps.emails.signals')


@receiver(post_save, sender=EmailAddress)
def email_addresses_updated(sender, instance, created, **kwargs):
    """
    Manage email model updates, only one primary/verified email per user
    """

    if kwargs['raw']:
        # https://github.com/django/django/commit/18a2fb19074ce6789639b62710c279a711dabf97
        return

    if instance.primary and instance.verified:
        user = instance.user

        # deleting all other user email addresses
        for email_address in user.emailaddress_set.exclude(pk=instance.pk).all():
            logger.info('Email "{}" primary and verified, deleting "{}"'.format(instance, email_address))

            email_address.delete()

    if not instance.verified:
        user = instance.user

        # deleting all other user email addresses
        for email_address in user.emailaddress_set.exclude(pk=instance.pk).filter(verified=False):
            logger.info('Email "{}" not verified, deleting "{}"'.format(instance, email_address))

            email_address.delete()



from django_sso_app.core.tests.factories import UserTestCase


class UserModelTestCase(UserTestCase):

    def test_django_superuser_user_creation_creates_sso_app_profile_with_sso_id_equals_to_username(self):
        admin_user_username = self._get_random_username()
        admin_user = self._get_new_admin_user(username=admin_user_username)

        self.assertEqual(admin_user.sso_id, admin_user.username)

    def test_django_staff_user_creation_creates_sso_app_profile_with_sso_id_equals_to_username(self):
        staff_user_username = self._get_random_username()
        staff_user = self._get_new_staff_user(username=staff_user_username)

        self.assertEqual(staff_user.sso_id, staff_user.username)



import sys

from .... import app_settings


signals_enabled = True
for c in app_settings.SIGNALS_DISABLED_COMMANDS:
    if c in sys.argv:
        signals_enabled = False
        break

if signals_enabled:
    if app_settings.BACKEND_ENABLED:
        from .backend import *
    else:
        from .app import *



import logging

from django.contrib import auth
from django.urls import reverse
from django.http import HttpResponseRedirect
from django.utils.deprecation import MiddlewareMixin  # https://stackoverflow.com/questions/42232606/django
                                                      # -exception-middleware-typeerror-object-takes-no-parameters

from ...permissions import is_authenticated, is_django_staff
from ...utils import set_session_key
from ... import app_settings

logger = logging.getLogger('django_sso_app.core.middleware')


PROFILE_INCOMPLETE_ENABLED_PATHS = [
    reverse('javascript-catalog'),
    reverse('profile.complete'),
]

class DjangoSsoAppAuthenticationBaseMiddleware(MiddlewareMixin):
    """
    See django.contrib.auth.middleware.RemoteUserMiddleware.
    """

    # Name of request header to grab username from.  This will be the key as
    # used in the request.META dictionary, i.e. the normalization of headers to
    # all uppercase and the addition of "HTTP_" prefix apply.
    header = app_settings.APIGATEWAY_CONSUMER_USERNAME_HEADER
    # Username for anonymous user
    anonymous_username = app_settings.APIGATEWAY_ANONYMOUS_CONSUMER_USERNAME

    @staticmethod
    def _clear_response_jwt(request):
        set_session_key(request, '__dssoa__clear_response_jwt', True)

    @staticmethod
    def _remove_invalid_user(request):
        """
        Removes the current authenticated user in the request which is invalid.
        """

        if is_authenticated(request.user):
            logger.info('removing invalid user "{}"'.format(request.user))

            auth.logout(request)

    @staticmethod
    def _redirect_to_profile_complete_response(user):
        _url = '{}'.format(app_settings.PROFILE_COMPLETE_URL)

        logger.info('User {} must complete profile,'
                    ' redirecting to {} ...'
                    .format(user, _url))

        response = HttpResponseRedirect(redirect_to=_url)

        return response

    @staticmethod
    def _request_path_is_disabled_for_incomplete_users(request_path):
        return (request_path not in PROFILE_INCOMPLETE_ENABLED_PATHS) and \
                not (request_path.startswith('/api/v1/')) and \
                not (request_path.startswith('/static/')) and \
                not (request_path.startswith('/media/')) and \
                not (request_path.startswith('/logout/')) and \
                not (request_path.startswith('/password/reset/')) and \
                not (request_path.startswith('/confirm-email/'))

    def process_request(self, request):
        raise NotImplementedError('process_request')

    def process_response(self, request, response):
        raise NotImplementedError('process_response')



"""
With these settings, tests run faster.
"""

from .base import *  # noqa
from .base import env

# GENERAL
# ------------------------------------------------------------------------------
# https://docs.djangoproject.com/en/dev/ref/settings/#debug
DEBUG = False
# https://docs.djangoproject.com/en/dev/ref/settings/#secret-key
SECRET_KEY = env(
    "DJANGO_SECRET_KEY",
    default="PpfDkCrcm2iJ6Mr9nkvfp6xbyT9w8DImXMi3H0vhRwVZMdeCokJjEmKCqJM89a4I",
)
# https://docs.djangoproject.com/en/dev/ref/settings/#test-runner
TEST_RUNNER = "django.test.runner.DiscoverRunner"

# CACHES
# ------------------------------------------------------------------------------
# https://docs.djangoproject.com/en/dev/ref/settings/#caches
CACHES = {
    "default": {
        "BACKEND": "django.core.cache.backends.locmem.LocMemCache",
        "LOCATION": "",
    }
}

# PASSWORDS
# ------------------------------------------------------------------------------
# https://docs.djangoproject.com/en/dev/ref/settings/#password-hashers
# PASSWORD_HASHERS = ["django.contrib.auth.hashers.MD5PasswordHasher"]  # django-sso-app

# TEMPLATES
# ------------------------------------------------------------------------------
TEMPLATES[0]["OPTIONS"]["loaders"] = [  # noqa F405
    (
        "django.template.loaders.cached.Loader",
        [
            "django.template.loaders.filesystem.Loader",
            "django.template.loaders.app_directories.Loader",
        ],
    )
]

# EMAIL
# ------------------------------------------------------------------------------
# https://docs.djangoproject.com/en/dev/ref/settings/#email-backend
EMAIL_BACKEND = "django.core.mail.backends.locmem.EmailBackend"

# Your stuff...
# ------------------------------------------------------------------------------



import logging


from django.contrib.auth import get_user_model
from django.db import IntegrityError
from django.forms import ValidationError
from django.http import Http404
# from django.utils.decorators import method_decorator
from django.utils.translation import ugettext_lazy as _
# from django.views.decorators.cache import cache_page
# from django.db.models import Q
from django.core.exceptions import SuspiciousOperation

from django_filters.rest_framework import DjangoFilterBackend

from allauth.account.adapter import get_adapter
from allauth.account.utils import filter_users_by_username, filter_users_by_email

from rest_framework import generics
from rest_framework import mixins
from rest_framework import status
from rest_framework import viewsets
from rest_framework import permissions
from rest_framework.response import Response
from rest_framework.views import APIView

from .serializers import (CheckUserExistenceSerializer,
                          UserSerializer, NewUserSerializer,
                          UserRevisionSerializer)

from ...apps.emails.models import EmailAddress
from ...permissions import StaffPermission
from ...permissions import is_staff, is_django_staff
from ...utils import invalidate_cookie, get_session_key
from ... import app_settings
from .filters import UserFilter

User = get_user_model()

logger = logging.getLogger('django_sso_app.core.apps.users')


class CheckUserExistenceApiView(APIView):
    """
    Retrieve a "condensed" user instance {"id": ".."} by either "username" or "email" as query params.
    """

    permission_classes = (permissions.AllowAny, )

    def get_object(self, login=None, username=None, email=None):
        try:
            if login is not None:
                #! NOQA
                email_users = filter_users_by_email(login)
                if len(email_users):
                    return email_users[0]

                username_users = filter_users_by_username(login)
                if len(username_users):
                    return username_users[0]

            elif username is None and email is not None:
                users = filter_users_by_email(email)
                if len(users):
                    return users[0]

            elif username is not None and email is None:
                users = filter_users_by_username(username)
                if len(users):
                    return users[0]

            raise User.DoesNotExist('Either login username or email must be set.')

        except (User.DoesNotExist, EmailAddress.DoesNotExist):
            return None

    def get(self, *args, **kwargs):
        login = self.request.query_params.get('login', self.request.GET.get('login', None))
        username = self.request.query_params.get('username', None)
        email = self.request.query_params.get('email', None)

        found = False
        if login is not None or username is not None or email is not None:
            item = self.get_object(login, username, email)
            found = item is not None

        if not found:
            raise Http404

        serializer = CheckUserExistenceSerializer(item)

        return Response(serializer.data)


class UserApiViewSet(mixins.ListModelMixin,
                     mixins.UpdateModelMixin,
                     mixins.CreateModelMixin,
                     mixins.RetrieveModelMixin,
                     mixins.DestroyModelMixin,
                     viewsets.GenericViewSet):
    """
    User api viewset
    """

    lookup_field = 'sso_app_profile__sso_id'
    # serializer_class = UserSerializer
    permission_classes = (permissions.IsAuthenticated,)
    filter_backends = (DjangoFilterBackend,)
    filterset_class = UserFilter

    def get_serializer_class(self):
        if self.request.method == 'POST':
            return NewUserSerializer
        else:
            return UserSerializer

    def get_queryset(self):
        user = self.request.user
        queryset = User.objects.filter(is_superuser=False, is_staff=False)

        if not is_staff(user):
            profile = user.get_sso_app_profile()
            queryset = queryset.filter(sso_app_profile__sso_id=profile.sso_id, is_active=True)

        return queryset

    def get_object(self, *args, **kwargs):
        sso_id = self.kwargs.get('sso_id')

        try:
            user = User.objects.get(sso_app_profile__sso_id=sso_id)
            self.check_object_permissions(self.request, user)
            return user
        except User.DoesNotExist:
            raise Http404

    def list(self, request, *args, **kwargs):
        """
        List users
        """
        logger.info('List users')

        return super(UserApiViewSet, self).list(request, *args, **kwargs)

    def create(self, request, *args, **kwargs):
        """
        Create user
        """
        logger.info('Creating user')

        if not is_staff(request.user):
            msg = 'Non staff user "{}" tried to create new user'.format(request.user)
            logger.warning(msg)

            raise(SuspiciousOperation(msg))

        return super(UserApiViewSet, self).create(request, *args, **kwargs)

    def partial_update(self, request, sso_id, *args, **kwargs):
        logger.info('Updating user')

        try:
            user = self.get_object()
            response = super(UserApiViewSet, self).partial_update(request, sso_id, *args, **kwargs)

        except (IntegrityError, ValidationError) as e:
            logger.exception('partial_update')

            logger.info('User {0} tried to update email with already registered one'.format(request.user))

            response = Response(_('Email already registered.'), status=status.HTTP_409_CONFLICT)

        response_status_code = response.status_code
        logger.debug('Partial_update status code is {}, {}'.format(response_status_code, request.data))

        if response_status_code == 200:
            adapter = get_adapter(request)
            requesting_user = request.user
            is_same_user = requesting_user == user

            if is_same_user and not is_django_staff(requesting_user):

                # must login again if password updated
                if get_session_key(request, '__dssoa__user_password_updated', False):
                    invalidate_cookie(response, app_settings.JWT_COOKIE_NAME)
                else:
                    adapter.try_update_response_jwt(user, response)

        return response


class UserRevisionsApiView(mixins.ListModelMixin,
                           viewsets.GenericViewSet):
    """
    User revisions entrypoint
    """

    queryset = User.objects.filter(is_superuser=False, is_staff=False).order_by('id')
    serializer_class = UserRevisionSerializer
    permission_classes = (StaffPermission,)

    def list(self, request, *args, **kwargs):
        """
        List users revisions
        """
        logger.info('List users revisions')
        return super(UserRevisionsApiView, self).list(request, *args, **kwargs)


class UserDetailApiView(generics.RetrieveAPIView):
    """
    User detail entrypoint
    """

    serializer_class = UserSerializer

    def get_queryset(self):
        return User.objects.all()

    def get_object(self, *args, **kwargs):
        user = getattr(self.request, 'user', None)
        return user



from django.contrib.admin import StackedInline
from django.contrib.auth import admin as auth_admin
from django.utils.translation import gettext, gettext_lazy as _

from .forms import UserChangeForm, UserCreationForm
from ..profiles.models import Profile

_list_display = ['username', 'email', 'is_staff', 'is_superuser', 'last_login', 'date_joined']
_search_fields = ['username', 'email']


class ProfileInline(StackedInline):
    model = Profile
    max_num = 1
    can_delete = False

    def get_readonly_fields(self, request, obj=None):
        # if obj is None:
        return ['sso_id', 'sso_rev']


class UserAdmin(auth_admin.UserAdmin):
    form = UserChangeForm
    add_form = UserCreationForm

    # pai
    # fieldsets = (('User', {'fields': ('name',)}),) + auth_admin.UserAdmin.fieldsets
    fieldsets = auth_admin.UserAdmin.fieldsets
    add_fieldsets = auth_admin.UserAdmin.add_fieldsets + (
        (_('Permissions'), {
            'fields': ('is_staff', 'is_superuser'),
        }),
    )

    list_display = _list_display
    search_fields = _search_fields

    # pai
    inlines = [ProfileInline]



from django.core.management.base import BaseCommand, CommandError
from django.contrib.auth import get_user_model

from allauth.account.adapter import get_adapter

User = get_user_model()


class Command(BaseCommand):
    help = 'Deletes all user devices (and jwts if APIGATEWAY_ENABLED)'

    def add_arguments(self, parser):
        parser.add_argument('username', nargs='+', type=str)

    def handle(self, *args, **options):
        usernames = options['username']
        users_count = len(usernames)
        updated_users = 0
        devices_count = 0
        deleted_devices = 0
        parsed_users = 0

        for username in usernames:
            try:
                user = User.objects.get(username=username)
            except User.DoesNotExist:
                self.stdout.write(self.style.ERROR('User "{}" does not exist'.format(username)))
                continue

            adapter = get_adapter()
            profile_devices_count = user.sso_app_profile.devices.count()
            devices_count += profile_devices_count
            deleted_profile_devices_count = adapter.remove_all_user_profile_devices(user)

            if deleted_profile_devices_count > 0:
                deleted_devices += deleted_profile_devices_count
                updated_users += 1

                self.stdout.write(
                    self.style.SUCCESS('Deleted {}/{} profile devices for "{}"'.format(profile_devices_count,
                                                                                       deleted_profile_devices_count,
                                                                                       user)))
            parsed_users += 1
            self.stdout.write(self.style.SUCCESS('{}/{}'.format(parsed_users, users_count)))

        self.stdout.write(self.style.SUCCESS('Updated {}/{} users'.format(updated_users, users_count)))
        self.stdout.write(self.style.SUCCESS('Deleted {}/{} devices'.format(deleted_devices, devices_count)))



import logging

from rest_framework import serializers
from rest_framework.reverse import reverse

from ...serializers import AbsoluteUrlSerializer, PartialObjectSerializer
from ... import app_settings
from .models import Profile

logger = logging.getLogger('django_sso_app.core.apps.profiles')


class ProfileSerializer(AbsoluteUrlSerializer):
    user = serializers.SerializerMethodField(required=False)
    sso_id = serializers.CharField()
    sso_rev = serializers.IntegerField()
    subscriptions = serializers.SerializerMethodField(required=False)

    ssn = serializers.CharField(required='ssn' in app_settings.REQUIRED_PROFILE_FIELDS)
    phone = serializers.CharField(required='phone' in app_settings.REQUIRED_PROFILE_FIELDS)

    first_name = serializers.CharField(required='first_name' in app_settings.REQUIRED_PROFILE_FIELDS)
    last_name = serializers.CharField(required='last_name' in app_settings.REQUIRED_PROFILE_FIELDS,
                                      allow_null=True,
                                      allow_blank=True)

    description = serializers.CharField(required='description' in app_settings.REQUIRED_PROFILE_FIELDS,
                                        allow_null=True, allow_blank=True)
    picture = serializers.CharField(required='picture' in app_settings.REQUIRED_PROFILE_FIELDS, allow_null=True)
    birthdate = serializers.DateField(required='birthdate' in app_settings.REQUIRED_PROFILE_FIELDS,
                                      allow_null=True)

    latitude = serializers.FloatField(required='latitude' in app_settings.REQUIRED_PROFILE_FIELDS)
    longitude = serializers.FloatField(required='longitude' in app_settings.REQUIRED_PROFILE_FIELDS)

    country = serializers.CharField(required='country' in app_settings.REQUIRED_PROFILE_FIELDS)
    address = serializers.CharField(required='address' in app_settings.REQUIRED_PROFILE_FIELDS, allow_null=True)

    language = serializers.CharField(required='language' in app_settings.REQUIRED_PROFILE_FIELDS,
                                     allow_null=True)

    alias = serializers.CharField(required='alias' in app_settings.REQUIRED_PROFILE_FIELDS,
                                  allow_null=True)

    # role = serializers.IntegerField(required='role' in app_settings.REQUIRED_PROFILE_FIELDS)

    groups = serializers.SerializerMethodField(required=False)

    meta = serializers.SerializerMethodField(required=False)

    is_incomplete = serializers.SerializerMethodField(required=False)
    required_fields = serializers.SerializerMethodField(required=False)
    completed_at = serializers.BooleanField(required=False)

    class Meta:
        model = Profile
        read_only_fields = (
            'url',
            'created_at',
            'sso_id', 'sso_rev',
            'subscriptions',
            'is_unsubscribed',
            'username', 'email',
            'user',
            'groups',
            'meta',
            'is_incomplete', 'required_fields', 'completed_at') + app_settings.USER_FIELDS
        fields = read_only_fields + app_settings.PROFILE_FIELDS

    def get_groups(self, instance):
        groups = list(instance.groups.values_list('name', flat=True))

        return groups

    def get_subscriptions(self, instance):
        serialized = []
        if app_settings.BACKEND_ENABLED:
            for el in instance.subscriptions.all():
                serialized.append({
                    'service_url': el.service.service_url,
                    'is_unsubscribed': getattr(el, 'is_unsubscribed', False)
                })
        return serialized

    def get_is_incomplete(self, instance):
        return instance.is_incomplete

    def get_required_fields(self, instance):
        serialized = []
        if app_settings.BACKEND_ENABLED:
            if instance.is_incomplete:
                for field in app_settings.REQUIRED_PROFILE_FIELDS:
                    if getattr(instance, field, None) in [None, '']:
                        serialized.append(field)

        return serialized

    def get_absolute_rest_url(self, obj):
        if getattr(obj, 'id', None) is not None:
            request = self.context['request']
            reverse_url = reverse('django_sso_app_profile:rest-detail', args=[obj.sso_id])

            return request.build_absolute_uri(reverse_url)

    def get_user(self, instance):
        try:
            user = instance.user
            request = self.context['request']
            return request.build_absolute_uri(user.get_absolute_rest_url())
        except:
            logger.warning('No profile.user url for {}.'.format(instance))

    def get_meta(self, instance):
        return instance.meta


class ProfilePublicSerializer(ProfileSerializer, PartialObjectSerializer):

    class Meta:
        model = Profile
        read_only_fields = (
            'url',
            'sso_id',
            'created_at', 'username', 'picture',
            'country',
            '_partial')
        fields = read_only_fields


class ProfileUnsubscriptionSerializer(serializers.Serializer):
    password = serializers.CharField()



import logging

from django.db.models.signals import pre_save
from django.dispatch import receiver

from .models import Status

logger = logging.getLogger('django_sso_app.core.apps.status')


@receiver(pre_save, sender=Status)
def state_pre_save(sender, instance, **kwargs):
    """
    Status model has been updated
    """

    if instance._state.adding:
        logger.debug('Status updated, syncing..')



import logging

from django.db import models
from django.utils.translation import ugettext_lazy as _
from django.urls import reverse

from ..groups.models import Group
from ..profiles.models import Profile
from ...models import CreatedAtModel, UpdatableModel, PublicableModel, DeactivableModel

logger = logging.getLogger('django_sso_app.core.apps.services')


class Service(CreatedAtModel, UpdatableModel, PublicableModel, DeactivableModel):
    """
    Protected service (site)
    """

    class Meta:
        app_label = 'django_sso_app'
        verbose_name = _('Service')

    name = models.CharField(_('name'), max_length=32, unique=True, null=True)
    service_url = models.CharField(_('service url'), max_length=255, unique=True, db_index=True)
    picture = models.TextField(_('picture'), null=True, blank=True)

    cookie_domain = models.CharField(_('cookie domain'), max_length=255, null=True, blank=True)  # unused but useful
    redirect_wait = models.PositiveIntegerField(_('redirect wait'), default=2000)

    subscription_required = models.BooleanField(_('subscription required'), default=False)
    is_public = models.BooleanField(_('is public'), default=True)

    groups = models.ManyToManyField(Group, verbose_name=_('groups'), blank=True,
                                    related_name='sso_app_services', related_query_name='sso_app_service')

    def get_tos(self, language):
        logger.debug('Service {0}, gettings tos for language {1}'.format(self, language))
        return self.terms_of_service.filter(language=language).first()

    def __str__(self):
        return '{}:{}'.format(self.name, self.service_url)

    def get_absolute_rest_url(self):
        return reverse("django_sso_app_service:rest-detail", args=[self.pk])


class TOS(CreatedAtModel, UpdatableModel, DeactivableModel):
    """
    Service TOS
    """
    class Meta:
        app_label = 'django_sso_app'
        verbose_name = _('TOS')

    language = models.CharField(_('language'), max_length=35)
    text = models.TextField(_('text'), default='')

    service = models.ForeignKey(Service, related_name="terms_of_service", on_delete=models.CASCADE,
                                verbose_name=_('service'))

    def __str__(self):
        return '{}:{}'.format(self.id, self.language)


class Subscription(CreatedAtModel, UpdatableModel, DeactivableModel, PublicableModel):
    """
    Service subscription
    """
    class Meta:
        app_label = 'django_sso_app'
        verbose_name = _('Subscription')

    profile = models.ForeignKey(
        Profile,
        to_field='sso_id',
        on_delete=models.CASCADE,
        related_name="subscriptions",
        verbose_name=_('profile')
    )
    service = models.ForeignKey(
        Service,
        on_delete=models.CASCADE,
        related_name="subscriptions",
        verbose_name=_('service')
    )
    unsubscribed_at = models.DateTimeField(_('unsubscribed at'), null=True, blank=True)

    @property
    def is_unsubscribed(self):
        return self.unsubscribed_at is not None

    @property
    def is_subscribed(self):
        return self.unsubscribed_at is None

    @property
    def user(self):
        return self.profile.user

    def __str__(self):
        return '{}@{}'.format(self.profile, self.service.service_url)

    def get_absolute_rest_url(self):
        return reverse("django_sso_app_service:rest-subscription-detail", args=[self.pk])



import logging

# import pyximport
# pyximport.install()

from ...functions import lists_differs

logger = logging.getLogger('django_sso_app.core.apps.groups')


def update_profile_groups(profile, groups_list):
    previous_groups_list = list(profile.groups.order_by('name').values_list('name', flat=True))

    if lists_differs(previous_groups_list, groups_list):
        logger.info('groups updated for "{}"'.format(profile))

        for group_name in groups_list:
            profile.add_to_group(group_name)



from django.contrib.auth import get_user_model

from django.urls import reverse

from rest_framework import status
from django_sso_app.core import app_settings
from django_sso_app.core.tests.factories import UserTestCase, responses


from django_sso_app.core.apps.profiles.models import Profile

User = get_user_model()


class TestApp(UserTestCase):

    @responses.activate
    def test_can_replicate_remote_profile(self):

        with self.settings(DJANGO_SSO_APP_SHAPE='app_persistence',
                           DJANGO_SSO_APP_SERVICE_URL='http://example.com'):

            remote_profile_uuid = self._get_random_uuid()

            remote_user_group_name = self._get_random_string()

            remote_user_object = self._get_remote_user_object(uuid=remote_profile_uuid,
                                                              service_name='example.com',
                                                              group_name=remote_user_group_name)

            mocked_url = app_settings.REMOTE_USER_URL.format(sso_id=remote_profile_uuid)
            profile_url = reverse('django_sso_app_profile:rest-detail', args=(remote_profile_uuid,))

            self._set_mocked_response(mocked_url, remote_user_object)

            client = self._get_client()
            client.cookies = self._get_valid_jwt_cookie(remote_profile_uuid)

            response = client.get(
                profile_url,
                content_type='application/json'
            )

            self.assertEqual(response.status_code, status.HTTP_200_OK)
            self.assertEqual(response.data.get('sso_id'), remote_profile_uuid, 'sso_id differs from remote and local')

            created_profile = Profile.objects.filter(sso_id=remote_profile_uuid).first()

            self.assertEqual(created_profile.groups.count(), 1, 'no profile groups created')
            self.assertEqual(created_profile.groups.filter(name=remote_user_group_name).count(), 1,
                             'users group missing')

    @responses.activate
    def test_can_replicate_remote_profile_with_apigateway(self):

        with self.settings(DJANGO_SSO_APP_SHAPE='app_persistence_apigateway',
                           DJANGO_SSO_APP_SERVICE_URL='http://example.com'):

            remote_profile_uuid = self._get_random_uuid()

            remote_user_group_name = self._get_random_string()

            remote_user_object = self._get_remote_user_object(uuid=remote_profile_uuid,
                                                              service_name='example.com',
                                                              group_name=remote_user_group_name)

            mocked_url = app_settings.REMOTE_USER_URL.format(sso_id=remote_profile_uuid)
            profile_url = reverse('django_sso_app_profile:rest-detail', args=(remote_profile_uuid,))

            self._set_mocked_response(mocked_url, remote_user_object)

            client = self._get_client()
            client.cookies = self._get_valid_jwt_cookie(remote_profile_uuid)

            response = client.get(
                profile_url,
                content_type='application/json',
                HTTP_X_CONSUMER_USERNAME=remote_profile_uuid.replace('-', '_'),
                HTTP_X_CONSUMER_GROUPS=remote_user_group_name
            )

            self.assertEqual(response.status_code, status.HTTP_200_OK)
            self.assertEqual(response.data.get('sso_id'), remote_profile_uuid, 'sso_id differs from remote and local')

            created_profile = Profile.objects.filter(sso_id=remote_profile_uuid).first()

            self.assertEqual(created_profile.groups.count(), 1, 'no profile groups created')
            self.assertEqual(created_profile.groups.filter(name=remote_user_group_name).count(), 1,
                             'users group missing')

    def test_can_login_by_apigateway_header(self):

        with self.settings(DJANGO_SSO_APP_SHAPE='app_apigateway'):

            remote_profile_uuid = self._get_random_uuid()
            remote_user_group_name = self._get_random_string()

            profile_url = reverse('django_sso_app_profile:rest-detail', args=(remote_profile_uuid,))

            client = self._get_client()
            client.cookies = self._get_valid_jwt_cookie(remote_profile_uuid)

            response = client.get(
                profile_url,
                content_type='application/json',
                HTTP_X_CONSUMER_USERNAME=remote_profile_uuid.replace('-', '_'),
                HTTP_X_CONSUMER_GROUPS=remote_user_group_name
            )

            self.assertEqual(response.status_code, status.HTTP_200_OK)
            self.assertEqual(response.data.get('sso_id'), remote_profile_uuid, 'sso_id differs from header')

            created_profile = Profile.objects.get(sso_id=remote_profile_uuid)

            self.assertEqual(Profile.objects.filter(sso_id=remote_profile_uuid).count(), 1, 'no new user created')
            self.assertEqual(created_profile.groups.count(), 1, 'group not created')
            self.assertEqual(created_profile.groups.filter(name=remote_user_group_name).count(), 1,
                             'group not equals header defined')

    @responses.activate
    def test_redirect_to_service_tos_if_service_not_subscribed(self):
        BACKEND_URL = 'http://accounts.example.com'
        SERVICE_URL = 'http://example.com'

        with self.settings(DJANGO_SSO_APP_SHAPE='app_persistence',
                           DJANGO_SSO_APP_BACKEND_URL=BACKEND_URL,
                           DJANGO_SSO_APP_SERVICE_URL=SERVICE_URL):
            # deleting pre generated user
            remote_profile_uuid = self._get_random_uuid()

            remote_user_group_name = self._get_random_string()

            remote_user_object = self._get_remote_user_object(uuid=remote_profile_uuid,
                                                              service_name='example.com',
                                                              group_name=remote_user_group_name,
                                                              service_subscribed=False)

            mocked_url = app_settings.REMOTE_USER_URL.format(sso_id=remote_profile_uuid)
            profile_url = reverse('django_sso_app_profile:rest-detail', args=(remote_profile_uuid,))

            self._set_mocked_response(mocked_url, remote_user_object)

            client = self._get_client()
            client.cookies = self._get_valid_jwt_cookie(remote_profile_uuid)

            response = client.get(
                profile_url,
                content_type='application/json'
            )

            self.assertEqual(response.status_code, status.HTTP_302_FOUND)
            self.assertEqual(response.url, BACKEND_URL + reverse('account_login') + '?next=http%3A%2F%2Fexample.com')

            created_profile = Profile.objects.filter(sso_id=remote_profile_uuid).first()

            self.assertEqual(created_profile.groups.count(), 1, 'no profile groups created')
            self.assertEqual(created_profile.groups.filter(name=remote_user_group_name).count(), 1,
                             'users group missing')

    @responses.activate
    def test_redirect_to_service_tos_if_service_not_subscribed_apigateway(self):
        BACKEND_URL = 'http://accounts.example.com'
        SERVICE_URL = 'http://example.com'

        with self.settings(DJANGO_SSO_APP_SHAPE='app_persistence',
                           DJANGO_SSO_APP_BACKEND_URL=BACKEND_URL,
                           DJANGO_SSO_APP_SERVICE_URL=SERVICE_URL):
            # deleting pre generated user
            remote_profile_uuid = self._get_random_uuid()

            remote_user_group_name = self._get_random_string()

            remote_user_object = self._get_remote_user_object(uuid=remote_profile_uuid,
                                                              service_name='example.com',
                                                              group_name=remote_user_group_name,
                                                              service_subscribed=False)

            mocked_url = app_settings.REMOTE_USER_URL.format(sso_id=remote_profile_uuid)
            profile_url = reverse('django_sso_app_profile:rest-detail', args=(remote_profile_uuid,))

            self._set_mocked_response(mocked_url, remote_user_object)

            client = self._get_client()
            client.cookies = self._get_valid_jwt_cookie(remote_profile_uuid)

            response = client.get(
                profile_url,
                content_type='application/json'
            )

            self.assertEqual(response.status_code, status.HTTP_302_FOUND)
            self.assertEqual(response.url, BACKEND_URL + reverse('account_login') + '?next=http%3A%2F%2Fexample.com')

            created_profile = Profile.objects.filter(sso_id=remote_profile_uuid).first()

            self.assertEqual(created_profile.groups.count(), 1, 'no profile groups created')
            self.assertEqual(created_profile.groups.filter(name=remote_user_group_name).count(), 1,
                             'users group missing')

    @responses.activate
    def test_redirect_to_profile_complete_if_profile_is_incomplete(self):
        BACKEND_URL = 'http://accounts.example.com'
        SERVICE_URL = 'http://example.com'

        with self.settings(DJANGO_SSO_APP_SHAPE='app_persistence',
                           DJANGO_SSO_APP_BACKEND_URL=BACKEND_URL,
                           DJANGO_SSO_APP_SERVICE_URL=SERVICE_URL):

            remote_profile_uuid = self._get_random_uuid()

            remote_user_group_name = self._get_random_string()

            remote_user_object = self._get_remote_user_object(uuid=remote_profile_uuid,
                                                              service_name='example.com',
                                                              group_name=remote_user_group_name,
                                                              service_subscribed=True,
                                                              incomplete=True)

            mocked_url = app_settings.REMOTE_USER_URL.format(sso_id=remote_profile_uuid)
            profile_url = reverse('django_sso_app_profile:rest-detail', args=(remote_profile_uuid,))

            self._set_mocked_response(mocked_url, remote_user_object)

            client = self._get_client()
            client.cookies = self._get_valid_jwt_cookie(remote_profile_uuid)

            response = client.get(
                profile_url,
                content_type='application/json'
            )

            self.assertEqual(response.status_code, status.HTTP_302_FOUND)
            self.assertEqual(response.url, BACKEND_URL + '/profile/complete/' + '?next=http%3A%2F%2Fexample.com')

            created_profile = Profile.objects.filter(sso_id=remote_profile_uuid).first()

            self.assertEqual(created_profile, None, 'incomplete user created')

    @responses.activate
    def test_redirect_to_profile_complete_if_profile_is_incomplete_apigateway(self):
        BACKEND_URL = 'http://accounts.example.com'
        SERVICE_URL = 'http://example.com'

        with self.settings(DJANGO_SSO_APP_SHAPE='app_persistence_apigateway',
                           DJANGO_SSO_APP_BACKEND_URL=BACKEND_URL,
                           DJANGO_SSO_APP_SERVICE_URL=SERVICE_URL):
            # deleting pre generated user
            remote_profile_uuid = self._get_random_uuid()

            remote_user_group_name = self._get_random_string()

            remote_user_object = self._get_remote_user_object(uuid=remote_profile_uuid,
                                                              service_name='example.com',
                                                              group_name=remote_user_group_name,
                                                              service_subscribed=True,
                                                              incomplete=True)

            mocked_url = app_settings.REMOTE_USER_URL.format(sso_id=remote_profile_uuid)
            profile_url = reverse('django_sso_app_profile:rest-detail', args=(remote_profile_uuid,))

            self._set_mocked_response(mocked_url, remote_user_object)

            client = self._get_client()
            client.cookies = self._get_valid_jwt_cookie(remote_profile_uuid)

            response = client.get(
                profile_url,
                content_type='application/json',
                HTTP_X_CONSUMER_USERNAME=remote_profile_uuid.replace('-', '_'),
                HTTP_X_CONSUMER_GROUPS=remote_user_group_name
            )

            self.assertEqual(response.status_code, status.HTTP_302_FOUND)
            self.assertEqual(response.url, BACKEND_URL + '/profile/complete/' + '?next=http%3A%2F%2Fexample.com')

            created_profile = Profile.objects.filter(sso_id=remote_profile_uuid).first()

            self.assertEqual(created_profile, None, 'incomplete user created')



import logging

from ... import app_settings

from .backend import DjangoSsoAppApiGatewayBackendAuthenticationBackend, DjangoSsoAppJwtBackendAuthenticationBackend
from .app import DjangoSsoAppApiGatewayAppAuthenticationBackend, DjangoSsoAppJwtAppAuthenticationBackend

logger = logging.getLogger('django_sso_app.core.authentication')


class DjangoSsoAppApiGatewayAuthenticationBackend(DjangoSsoAppApiGatewayBackendAuthenticationBackend,
                                                  DjangoSsoAppApiGatewayAppAuthenticationBackend):

    """
    Authenticates by request CONSUMER_USERNAME header

    See django.contrib.auth.backends.RemoteUserBackend.

    """

    backend_path = 'django_sso_app.core.authentication.backends.DjangoSsoAppApiGatewayAuthenticationBackend'

    def authenticate(self, request, consumer_username, encoded_jwt, decoded_jwt):
        if app_settings.BACKEND_ENABLED:
            return self.backend_authenticate(request,
                                             consumer_username,
                                             encoded_jwt,
                                             decoded_jwt)
        else:
            return self.app_authenticate(request,
                                         consumer_username,
                                         encoded_jwt,
                                         decoded_jwt)

class DjangoSsoAppJwtAuthenticationBackend(DjangoSsoAppJwtBackendAuthenticationBackend,
                                           DjangoSsoAppJwtAppAuthenticationBackend):
    """
    Authenticates by request jwt

    Example JWT:

        {
          "sso_id": "<uuid>",
          "sso_rev": 0,
          "id": 1,
          "fp": "b28493a6a29a5a38973a8a3e3abe7f34",
          "iss": "nxhnkMPKcpRWaTKwNOvGcLcs5MHJINOg"
        }

    """

    backend_path = 'django_sso_app.core.authentication.backends.DjangoSsoAppApiGatewayAuthenticationBackend'

    def authenticate(self, request, encoded_jwt, decoded_jwt):
        if app_settings.BACKEND_ENABLED:
            return self.backend_authenticate(request, encoded_jwt, decoded_jwt)
        else:
            return self.app_authenticate(request, encoded_jwt, decoded_jwt)



import logging

from django.core.exceptions import ObjectDoesNotExist
from django.utils.http import urlencode
from django.http import HttpResponseRedirect
from django.contrib.auth import get_user_model
from django.contrib.auth.backends import ModelBackend

from ...apps.users.utils import fetch_remote_user, create_local_user_from_remote_backend, \
                              update_local_user_from_remote_backend, create_local_user_from_jwt, \
                              create_local_user_from_apigateway_headers

from ...apps.profiles.models import Profile
from ...apps.groups.utils import update_profile_groups
from ...apps.users.utils import add_user_to_default_groups
from ...apps.api_gateway.functions import get_apigateway_sso_id, get_apigateway_profile_groups_from_header
from ...exceptions import ServiceSubscriptionRequiredException, ProfileIncompleteException
from ...utils import set_session_key
from ... import app_settings


logger = logging.getLogger('django_sso_app.core.authentication.backends')
User = get_user_model()


class DjangoSsoAppAppBaseAuthenticationBackend(ModelBackend):

    def try_replicate_user(self, request, sso_id, encoded_jwt, decoded_jwt):
        logger.debug('try_replicate_user')

        if app_settings.REPLICATE_PROFILE:
            logger.info('Replicate user with sso_id "{}" from remote backend'.format(sso_id))

            # create local profile from SSO
            backend_user = fetch_remote_user(sso_id=sso_id, encoded_jwt=encoded_jwt)
            backend_user_profile = backend_user['profile']

            if backend_user_profile.get('is_incomplete', False):
                self.redirect_to_profile_complete(request.user)

            user = create_local_user_from_remote_backend(backend_user)

        else:
            user = None

        return user

    def try_update_user(self, sso_id, user, user_profile, encoded_jwt, decoded_jwt):
        logger.debug('try_update_user')

        rev_changed = user_profile.sso_rev < decoded_jwt['sso_rev']
        first_access = not user.is_active and not user_profile.is_unsubscribed

        if rev_changed or first_access:
            if rev_changed:
                logger.info('Rev changed from "{}" to "{}" for user "{}", updating ...'
                            .format(user_profile.sso_rev, decoded_jwt['sso_rev'],
                                    user))
            if first_access:
                logger.info('"{} first access, updating ...'.format(user))

            # local profile updated from django_sso_app instance, do not update sso_rev
            setattr(user, '__dssoa__creating', True)

            remote_user = fetch_remote_user(sso_id=sso_id, encoded_jwt=encoded_jwt)
            user = update_local_user_from_remote_backend(remote_user=remote_user,
                                                         profile=user_profile)

            logger.info('{} updated with latest data from SSO'.format(user))

        else:
            logger.info('Nothing changed for apigateway consumer "{}"'.format(user))

        return user

    def try_update_profile_subscriptions(self, user):
        # Redirect user to Term Of Service.
        # is_to_subscribe = getattr(user, '__dssoa__is_to_subscribe', False)
        if user and app_settings.REPLICATE_PROFILE and app_settings.SERVICE_SUBSCRIPTION_REQUIRED:
            remote_user = getattr(user, '__dssoa__remote_user', None)

            if remote_user is not None:
                logger.info('Check profile subscriptions for "{}"'.format(app_settings.SERVICE_URL))

                remote_profile_subscriptions = remote_user['profile']['subscriptions']
                logger.debug('remote_profile_subscriptions "{}"'.format(remote_profile_subscriptions))

                active_subscriptions = []
                for s in remote_profile_subscriptions:
                    if not s['is_unsubscribed']:
                        active_subscriptions.append(s['service_url'])

                if app_settings.SERVICE_URL not in active_subscriptions:
                    self.redirect_to_service_subscription(user)

    def redirect_to_service_subscription(self, user):
        _qs = urlencode({'next': app_settings.SERVICE_URL})
        _url = '{}{}?{}'.format(app_settings.BACKEND_URL, app_settings.LOGIN_URL, _qs)

        logger.info('User {} must agree to the Terms of Service,'
                    ' redirecting to {} ...'
                    .format(user, _url))

        response = HttpResponseRedirect(redirect_to=_url)

        raise ServiceSubscriptionRequiredException(response)

    def redirect_to_profile_complete(self, user):
        _qs = urlencode({'next': app_settings.SERVICE_URL})
        _url = '{}{}?{}'.format(app_settings.BACKEND_URL, app_settings.PROFILE_COMPLETE_URL, _qs)

        logger.info('User {} must complete profile,'
                    ' redirecting to {} ...'
                    .format(user, _url))

        response = HttpResponseRedirect(redirect_to=_url)

        raise ProfileIncompleteException(response)


class DjangoSsoAppApiGatewayAppAuthenticationBackend(DjangoSsoAppAppBaseAuthenticationBackend):
    """
    Authenticates by request CONSUMER_USERNAME header
    """

    def try_replicate_user(self, request, sso_id, encoded_jwt, decoded_jwt):
        user = super(DjangoSsoAppApiGatewayAppAuthenticationBackend, self).try_replicate_user(request, sso_id, encoded_jwt, decoded_jwt)

        if user is None:
            logger.info('Creating user from headers')

            user = create_local_user_from_apigateway_headers(request)

        return user

    def try_update_profile_groups(self, request, user):
        if app_settings.MANAGE_USER_GROUPS:
            logger.info('Updating profile groups for "{}"'.format(user))

            consumer_groups_header = request.META.get(app_settings.APIGATEWAY_CONSUMER_GROUPS_HEADER, None)
            profile_groups = get_apigateway_profile_groups_from_header(consumer_groups_header)

            update_profile_groups(user.sso_app_profile, profile_groups)

    def app_authenticate(self, request, consumer_username, encoded_jwt, decoded_jwt):
        logger.info('APP authenticating by apigateway consumer {}'.format(consumer_username))

        try:
            sso_id = get_apigateway_sso_id(consumer_username)
            profile = Profile.objects.get(sso_id=sso_id)
            user = profile.user

        except ObjectDoesNotExist:
            logger.info('No profile with id "{}"'.format(sso_id))
            try:
                user = self.try_replicate_user(request, sso_id, encoded_jwt, decoded_jwt)

            except ProfileIncompleteException:
                raise

            except Exception:
                logger.exception('Can not replicate user')

                return

        else:
            if app_settings.REPLICATE_PROFILE:
                logger.debug('Must update profile')

                if decoded_jwt is None:
                    logger.warning('decoded_jwt not set')
                    return

                user = self.try_update_user(sso_id, user, profile, encoded_jwt, decoded_jwt)

        # update user profile relations
        self.try_update_profile_groups(request, user)
        self.try_update_profile_subscriptions(user)

        add_user_to_default_groups(user)

        set_session_key(request, '__dssoa__requesting_user', user)

        return user


class DjangoSsoAppJwtAppAuthenticationBackend(DjangoSsoAppAppBaseAuthenticationBackend):
    """
    Authenticates by request jwt
    """

    def try_replicate_user(self, request, sso_id, encoded_jwt, decoded_jwt):

        user = super(DjangoSsoAppJwtAppAuthenticationBackend, self).try_replicate_user(request, sso_id, encoded_jwt, decoded_jwt)

        if user is None:
            # create local profile from jwt
            logger.info('Replicating user with sso_id "{}" from JWT'.format(sso_id))

            user = create_local_user_from_jwt(decoded_jwt)

        return user

    def try_update_profile_groups(self, request, user):
        # Update user groups
        if app_settings.MANAGE_USER_GROUPS:
            if app_settings.REPLICATE_PROFILE:
                logger.info('try_update_profile_groups for "{}"'.format(user))

                remote_user_object = getattr(user, '__dssoa__remote_user', None)

                if remote_user_object is not None:
                    remote_user_object_groups = remote_user_object.get('groups', [])

                    update_profile_groups(user.sso_app_profile, remote_user_object_groups)

    def app_authenticate(self, request, encoded_jwt, decoded_jwt):
        logger.info('backend authenticating by request jwt')

        if encoded_jwt is None or decoded_jwt is None:
            logger.debug('request jwt not set, skipping authentication')

            return

        try:
            sso_id = decoded_jwt['sso_id']
            user_profile = Profile.objects.get(sso_id=sso_id)
            user = user_profile.user

            if app_settings.REPLICATE_PROFILE:
                logger.debug('try_update_user "{}" jwt consumer "{}"'.format(sso_id, sso_id))

                user = self.try_update_user(sso_id, user, user_profile, encoded_jwt, decoded_jwt)

            else:
                # just updates user groups
                logger.debug('Do not replicate profile')

        except ObjectDoesNotExist:
            try:
                user = self.try_replicate_user(request, sso_id, encoded_jwt, decoded_jwt)

            except ProfileIncompleteException:
                raise

            except Exception:
                logger.exception('cannot replicate remote user')
                return

        # update relations
        self.try_update_profile_groups(request, user)
        self.try_update_profile_subscriptions(user)

        add_user_to_default_groups(user)

        set_session_key(request, '__dssoa__requesting_user', user)

        return user



import logging

from django.core.exceptions import ImproperlyConfigured
from django.conf import settings
from django.urls import reverse
from django.utils.encoding import smart_str

from jwt.exceptions import InvalidSignatureError

from ... import app_settings
from ...utils import set_session_key, get_session_key
from ...permissions import is_authenticated, is_django_staff
from ...exceptions import RequestHasValidJwtWithNoDeviceAssociated
from ...tokens.utils import get_request_jwt, jwt_decode
from ...apps.api_gateway.functions import get_apigateway_sso_id

from .backend import DjangoSsoAppAuthenticationBackendMiddleware
from .app import DjangoSsoAppAuthenticationAppMiddleware

logger = logging.getLogger('django_sso_app.core.authentication.middleware')


ADMIN_URL = '/{}'.format(settings.ADMIN_URL)
SSO_ID_JWT_KEY = 'sso_id'
PROFILE_INCOMPLETE_ENABLED_PATHS = [
    reverse('javascript-catalog'),
    reverse('profile.complete'),
]


class DjangoSsoAppAuthenticationMiddleware(DjangoSsoAppAuthenticationBackendMiddleware,
                                           DjangoSsoAppAuthenticationAppMiddleware):
    """
    See django.contrib.auth.middleware.RemoteUserMiddleware.
    """

    def process_request(self, request):
        # AuthenticationMiddleware is required so that request.user exists.
        if not hasattr(request, 'user'):
            raise ImproperlyConfigured(
                "SSO middleware requires the authentication middleware to be"
                " installed.  Edit your MIDDLEWARE setting to insert"
                " 'django.contrib.auth.middleware.AuthenticationMiddleware'"
                " before the SsoMiddleware class.")

        request_path = request.path
        request_method = request.method
        request_ip = request.META.get('REMOTE_ADDR', None)
        request_user = request.user

        logger.info('--- "{}" request "{}" path "{}" method "{}" user "{}"'.format(request_ip,
                                                                                   id(request),
                                                                                   request_path,
                                                                                   request_method,
                                                                                   request_user))

        if is_authenticated(request_user):
            set_session_key(request, '__dssoa__requesting_user', request_user)

            if is_django_staff(request_user):
                logger.info('Skipping django staff user "{}"'.format(request_user))
                #  >= 1.10 has is_authenticated as parameter
                # If a staff user is already authenticated, we don't need to
                # continue

                return

            elif request_path.startswith(ADMIN_URL):
                logger.warning('Non staff user "{}" called admin path'.format(request_user))

                self._remove_invalid_user(request)
                self._clear_response_jwt(request)

                return

        # saving request info
        set_session_key(request, '__dssoa__request_ip', request_ip)

        """
        if self.request_path.startswith('/api/v1/passepartout'):
            logger.info('is passepartout path')
            # set_session_key(request, '__dssoa__is_passepartout_path', True)
        """

        try:
            request_jwt = get_request_jwt(request, encoded=False)

            if request_jwt is None:
                logger.debug('No JWT in request, skipping')

                return

            else:
                # decoding request JWT
                request_device, decoded_jwt = jwt_decode(request_jwt, verify=True)

        except KeyError:
            logger.exception('Malformed JWT "{}"'.format(request_jwt))

            self._remove_invalid_user(request)
            self._clear_response_jwt(request)

            return

        except InvalidSignatureError:
            logger.warning('Invalid JWT signature "{}"'.format(request_jwt))

            self._remove_invalid_user(request)
            self._clear_response_jwt(request)

            return

        except RequestHasValidJwtWithNoDeviceAssociated:
            logger.warning('RequestHasValidJwtWithNoDeviceAssociated "{}"'.format(request_jwt))

            self._remove_invalid_user(request)
            self._clear_response_jwt(request)

            return

        except Exception:
            logger.exception('Generic exception "{}"'.format(request_jwt))

            self._remove_invalid_user(request)
            self._clear_response_jwt(request)

            return

        apigateway_enabled = app_settings.APIGATEWAY_ENABLED
        request_jwt_sso_id = decoded_jwt[SSO_ID_JWT_KEY]

        if apigateway_enabled:
            # getting sso_id by apigateway consumer username
            try:
                consumer_username = request.META[self.header]

                if consumer_username == self.anonymous_username:
                    logger.info('consumer is anonymous, returning')

                    self._remove_invalid_user(request)

                    return

            except KeyError as e:
                # If specified header or jwt doesn't exist then remove any existing
                # authenticated user, or return (leaving request.user set to
                # AnonymousUser by the AuthenticationMiddleware).

                logger.info('Missing apigateway header "{}"'.format(e))

                self._remove_invalid_user(request)

                return

            else:
                sso_id = get_apigateway_sso_id(consumer_username)

                set_session_key(request, '__dssoa__apigateway__consumer_username', consumer_username)

        else:
            sso_id = request_jwt_sso_id

        # If the user is already authenticated and that user is active and the
        # one we are getting passed in the headers, then the correct user is
        # already persisted in the session and we don't need to continue.
        if is_authenticated(request_user):
            # double check sso_id equality between received jwt and sso_app_profile
            if smart_str(request_user.sso_id) != sso_id:

                # An authenticated user is associated with the request, but
                # it does not match the authorized user in the header.
                logger.warning('credentials and request_user sso_id differs! "{}" "{}"'.format(sso_id,
                                                                                               request_user.sso_id))

                self._remove_invalid_user(request)
                self._clear_response_jwt(request)

                return

        # authentication
        if app_settings.BACKEND_ENABLED:
            self.backend_process_request(request, request_jwt, decoded_jwt)
        else:
            self.app_process_request(request, request_jwt, decoded_jwt)

    def process_response(self, request, response):
        # getting request info
        requesting_user = get_session_key(request, '__dssoa__requesting_user')
        request_ip = get_session_key(request, '__dssoa__request_ip')

        if app_settings.BACKEND_ENABLED:
            response = self.backend_process_response(request, response)
        else:
            response = self.app_process_response(request, response)

        logger.info('--- "{}" request "{}" user "{}" path "{}" method "{}" ({})'.format(request_ip,
                                                                                        id(request),
                                                                                        requesting_user,
                                                                                        request.path,
                                                                                        request.method,
                                                                                        response.status_code))

        return response



from django.conf.urls import url
from rest_framework.urlpatterns import format_suffix_patterns

from .views import (CheckUserExistenceApiView,
                    UserApiViewSet,
                    UserRevisionsApiView,
                    UserDetailApiView)

_urlpatterns = [
    url(r'^$', UserApiViewSet.as_view({'get': 'list', 'post': 'create'}), name='rest-list'),

    url(r'^(?P<sso_id>[-_\w]+)/$',
        UserApiViewSet.as_view({'get': 'retrieve', 'patch': 'partial_update'},
                               name='rest-detail'),
        name='rest-detail'),
]

extra_urlpatterns = (format_suffix_patterns([
                         url(r'^user/$', UserDetailApiView.as_view(), name="user_detail"),
                         url(r'^check-user/$', CheckUserExistenceApiView.as_view(), name="check_user_existence"),
                         url(r'^get-revisions/$', UserRevisionsApiView.as_view({'get': 'list'}), name="get_users_revisions")]),
                     'django_sso_app_user_extra')

urlpatterns = (format_suffix_patterns(_urlpatterns), 'django_sso_app_user')



import json

import jwt
from django.contrib.auth import get_user_model
from django.test import Client
from django.urls import reverse
from rest_framework import status

from django_sso_app.core.tests.factories import UserTestCase

User = get_user_model()


class TestProfileRetrieve(UserTestCase):

    def test_user_can_retrieve_profile_with_subscriptions(self):
        """
          Registration via rest_auth view
        """
        new_pass = self._get_random_pass()
        new_user = self._get_new_subscribed_user(password=new_pass)
        profile = new_user.sso_app_profile

        client = self._get_client()
        response = client.post(
            reverse('rest_login'),
            data=json.dumps(self._get_login_object(new_user.email, new_pass)),
            content_type='application/json'
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

        response2 = client.get(
            reverse('django_sso_app_profile:rest-detail', args=(profile.sso_id,)),
            content_type='application/json'
        )

        self.assertEqual(response2.status_code, status.HTTP_200_OK)

        self.assertEqual(response2.data.get('subscriptions'),
                         [{'service_url': el.service.service_url, 'is_unsubscribed': el.is_unsubscribed} for el in
                          profile.subscriptions.all()])


class TestProfileUpdate(UserTestCase):

    def test_profile_update_refreshes_and_returns_current_device_jwt_and_deletes_other_devices(self):
        """
        Profile update refreshes caller jwt and deletes all but caller devices
        """
        new_pass = self._get_random_pass()
        new_user = self._get_new_user(password=new_pass)

        profile = new_user.sso_app_profile
        first_rev = profile.sso_rev

        login_object = self._get_login_object(new_user.email, new_pass)

        client = self._get_client()
        response = client.post(
            reverse('rest_login'),
            data=json.dumps(login_object),
            content_type='application/json'
        )

        cookies = response.cookies

        profile_update = self._get_new_profile_object()

        response2 = client.patch(
            reverse('django_sso_app_profile:rest-detail', args=[profile.sso_id]),
            data=json.dumps(profile_update),
            content_type='application/json'
        )

        cookies2 = response2.cookies

        profile.refresh_from_db()
        second_rev = profile.sso_rev

        self.assertEqual(second_rev, first_rev + 1, 'rev not incremented')

        self.assertNotEqual(self._get_response_jwt_cookie(response2), '', 'profile update do not return jwt cookie')

        devices = profile.devices.all()

        self.assertEqual(devices.count(), 1, 'other devices not deleted')
        self.assertEqual(devices.first().fingerprint, login_object['fingerprint'], 'device not updated')

        jwt_cookie = cookies.get('jwt').value
        jwt_cookie2 = cookies2.get('jwt').value

        # print('\nRISPOSTA, user rev', self.profile.sso_rev, '\n\nTOKEN', jwt_cookie)

        unverified_payload = jwt.decode(jwt_cookie, None, False)
        unverified_payload2 = jwt.decode(jwt_cookie2, None, False)

        self.assertEqual(unverified_payload2.get('sso_rev'), unverified_payload.get('sso_rev') + 1,
                         'rev not incremented in cookie')

        for f in profile_update:
            model_val = getattr(profile, f, None)
            object_val = profile_update.get(f, None)

            if f == 'birthdate':
                model_val = str(model_val)
            elif f == 'country':
                object_val = object_val.upper()
                model_val = str(model_val).upper()

            self.assertEqual(model_val, object_val)


class TestCompleteUnsubscription(UserTestCase):

    def test_user_can_completely_unsubscribe_deleting_devices_and_disables_login(self):
        """
        Profile update refreshes caller jwt and deletes all but caller devices
        """
        new_pass = self._get_random_pass()
        new_user = self._get_new_user(password=new_pass)

        profile = new_user.sso_app_profile
        first_rev = profile.sso_rev

        login_object = self._get_login_object(new_user.email, new_pass)

        client = self._get_client()
        response = client.post(
            reverse('rest_login'),
            data=json.dumps(login_object),
            content_type='application/json'
        )

        url2 = reverse('django_sso_app_profile:rest-complete-unsubscription', args=[profile.sso_id])

        response2 = client.post(
            url2,
            data=json.dumps({'password': new_pass}),
            content_type='application/json'
        )
        cookies2 = response2.cookies

        self.assertEqual(response2.status_code, status.HTTP_200_OK)

        profile.refresh_from_db()
        second_rev = profile.sso_rev
        self.assertEqual(second_rev, first_rev + 1, 'rev not incremented')

        self.assertTrue(len(cookies2.keys()) > 0, 'no jwt cookie set')

        devices = profile.devices.all()

        self.assertEqual(devices.count(), 0, 'devices not deleted')

        self.assertEqual(response2.data, 'Successfully completely unsubscribed.')

        client2 = self._get_client()
        response3 = client2.post(
            reverse('rest_login'),
            data=json.dumps(login_object),
            content_type='application/json'
        )

        self.assertEqual(response3.status_code, status.HTTP_400_BAD_REQUEST, 'user can login even if unsubscribed')


class TestProfileComplete(UserTestCase):

    def test_user_can_complete_profile_with_username_changed(self):
        """
          Registration via rest_auth view
        """
        new_pass = self._get_random_pass()
        new_incomplete_user = self._get_new_incomplete_user(password=new_pass)

        profile = new_incomplete_user.sso_app_profile
        first_rev = profile.sso_rev

        login_object = self._get_login_object(new_incomplete_user.email, new_pass)

        client = self._get_client()
        response = client.post(
            reverse('rest_login'),
            data=json.dumps(login_object),
            content_type='application/json'
        )

        cookies = response.cookies
        print('RESP', response.data)

        self.assertEqual(response.status_code, status.HTTP_200_OK)

        profile_complete = {
            'username': self._get_random_username(),
            'profile': self._get_new_profile_object()
        }

        response2 = client.patch(
            reverse('django_sso_app_user:rest-detail', args=[new_incomplete_user.sso_app_profile.sso_id]),
            data=json.dumps(profile_complete),
            content_type='application/json'
        )

        self.assertEqual(response2.status_code, status.HTTP_200_OK)

        new_incomplete_user.refresh_from_db()
        self.assertEqual(new_incomplete_user.username, profile_complete['username'])

        self.assertEqual(new_incomplete_user.sso_app_profile.first_name, profile_complete['profile']['first_name'])

        self.assertEqual(new_incomplete_user.sso_app_profile.is_incomplete, False)

    def test_user_cant_update_username_on_profile_complete(self):
        """
          Registration via rest_auth view
        """
        from django_sso_app import app_settings

        new_pass = self._get_random_pass()
        new_incomplete_user = self._get_new_incomplete_user(password=new_pass)

        profile = new_incomplete_user.sso_app_profile
        first_rev = profile.sso_rev

        login_object = self._get_login_object(new_incomplete_user.email, new_pass)

        client = self._get_client()
        response = client.post(
            reverse('rest_login'),
            data=json.dumps(login_object),
            content_type='application/json'
        )

        cookies = response.cookies
        print('RESP', response.data)

        self.assertEqual(response.status_code, status.HTTP_200_OK)

        profile_complete = self._get_new_profile_complete_object()

        response2 = client.patch(
            reverse('django_sso_app_user:rest-detail', args=[new_incomplete_user.sso_app_profile.sso_id]),
            data=json.dumps(profile_complete),
            content_type='application/json'
        )

        self.assertEqual(response2.status_code, status.HTTP_200_OK)

        new_incomplete_user.refresh_from_db()

        self.assertEqual(new_incomplete_user.username, profile_complete['username'])

        self.assertEqual(new_incomplete_user.sso_app_profile.first_name,
                         profile_complete['profile']['first_name'])

        self.assertEqual(new_incomplete_user.sso_app_profile.is_incomplete, False)

        invalid_profile_complete = self._get_new_profile_complete_object()

        response3 = client.patch(
            reverse('django_sso_app_user:rest-detail', args=[new_incomplete_user.sso_app_profile.sso_id]),
            data=json.dumps(invalid_profile_complete),
            content_type='application/json'
        )

        self.assertEqual(response3.status_code, status.HTTP_200_OK)

        new_incomplete_user.refresh_from_db()

        self.assertNotEqual(invalid_profile_complete['username'], new_incomplete_user.username)



from ..core.settings.common import *


DJANGO_SSO_APP_DJANGO_AUTHENTICATION_BACKENDS = [
    # requests
    'django_sso_app.core.authentication.backends.DjangoSsoAppApiGatewayAuthenticationBackend',
    'django_sso_app.core.authentication.backends.DjangoSsoAppJwtAuthenticationBackend'
]



import logging

from django.contrib.auth import get_user_model
from django.shortcuts import redirect
from django.urls import reverse
from django.urls import reverse_lazy
from django.views.generic import UpdateView
from django.contrib.auth.mixins import LoginRequiredMixin

from rest_framework import permissions
from rest_framework import status
from rest_framework import viewsets
from rest_framework.response import Response
from rest_framework import generics

from django_filters.rest_framework import DjangoFilterBackend

from allauth.account.adapter import get_adapter

from .forms import UserProfileForm
from .models import Profile
from .serializers import ProfileSerializer, ProfilePublicSerializer, ProfileUnsubscriptionSerializer
from ... import app_settings

from ...permissions import is_staff
from ...utils import invalidate_cookie
from ...mixins import WebpackBuiltTemplateViewMixin

from .filters import ProfileFilter

User = get_user_model()

logger = logging.getLogger('django_sso_app.core.apps.profiles')


class ProfileView(LoginRequiredMixin, WebpackBuiltTemplateViewMixin):
    template_name = "account/profile.html"

    def get_object(self, queryset=None):
        request_user = getattr(self.request, 'user', None)

        return request_user.get_sso_app_profile()

    def get_context_data(self, *args, **kwargs):
        context = super(ProfileView, self).get_context_data(*args, **kwargs)
        self.object = self.get_object()

        if self.object is not None:
            profile = self.object

            user_fields = []
            for field in app_settings.PROFILE_FIELDS:
                user_fields.append((field, getattr(profile, field)))

            context['user_fields'] = user_fields

        return context

    def get(self, request, *args, **kwargs):
        logger.info('Getting profile')

        try:
            return super(ProfileView, self).get(request, *args, **kwargs)

        except:
            logger.exception('Error getting profile')
            raise


class ProfileUpdateView(LoginRequiredMixin, UpdateView, WebpackBuiltTemplateViewMixin):
    model = Profile
    template_name = "account/profile_update.html"
    form_class = UserProfileForm

    success_url = reverse_lazy('profile')

    def get_object(self, queryset=None):
        request_user = getattr(self.request, 'user', None)

        return request_user.get_sso_app_profile()

    def get_context_data(self, *args, **kwargs):
        self.object = self.get_object()

        context = super(WebpackBuiltTemplateViewMixin, self).get_context_data(*args, **kwargs)
        return context

    def get(self, request, *args, **kwargs):
        logger.info('Getting profile update')

        try:
            profile = self.get_object()
            return super(ProfileUpdateView, self).get(request, profile.pk, *args, **kwargs)

        except:
            logger.exception('Error getting profile update')
            raise

    def post(self, request, *args, **kwargs):
        logger.info('Posting profile update')

        try:
            profile = self.get_object()
            response = super(ProfileUpdateView, self).post(request, profile.pk, *args, **kwargs)

        except:
            logger.exception('Error posting profile update')
            raise

        else:
            user = profile.user
            adapter = get_adapter(request)

            adapter.try_update_response_jwt(user, response)

            return response


class ProfileCompleteView(ProfileUpdateView):
    template_name = "account/profile_complete.html"

    def get(self, request, *args, **kwargs):
        logger.info('Getting profile complete')

        if not request.user.sso_app_profile.is_incomplete:
            return redirect(reverse('profile.update'))

        return super(ProfileCompleteView, self).get(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        logger.info('Posting profile create')

        if not request.user.sso_app_profile.is_incomplete:
            return redirect(reverse('profile.update'))

        return super(ProfileCompleteView, self).post(request, *args, **kwargs)


# api

class ProfileApiViewSet(viewsets.ModelViewSet):
    lookup_field = 'sso_id'
    permission_classes = (permissions.IsAuthenticated, )
    filter_backends = (DjangoFilterBackend,)
    filterset_class = ProfileFilter
    serializer_class = ProfileSerializer

    def get_queryset(self):
        requesting_user = self.request.user

        if is_staff(requesting_user):
            return Profile.objects.all()
        else:
            return Profile.objects.filter(user=requesting_user)

    def list(self, request, *args, **kwargs):
        """
        List profiles
        """
        return super(ProfileApiViewSet, self).list(request, *args, **kwargs)

    def retrieve(self, request, *args, **kwargs):
        """
        Profile detail
        """
        return super(ProfileApiViewSet, self).retrieve(request, *args, **kwargs)

    def partial_update(self, request, *args, **kwargs):
        """
        Profile update
        """
        response = super(ProfileApiViewSet, self).partial_update(request, *args, **kwargs)
        response_status_code = response.status_code
        logger.info('Profile partial_update status code is {}, {}'.format(response_status_code, request.data))

        if response_status_code == 200:
            profile = self.get_object()
            user = profile.user

            adapter = get_adapter(request)
            adapter.try_update_response_jwt(user, response)

        return response


class ProfileCompleteUnsubscriptionApiView(viewsets.GenericViewSet):
    """
    Completely unsubscribe user
    """

    permission_classes = (permissions.IsAuthenticated, )
    serializer_class = ProfileUnsubscriptionSerializer

    def get_serializer(self, *args, **kwargs):
        request = self.request
        serializer_class = self.get_serializer_class()

        return serializer_class(data=request.data, context={'request': request})

    def completely_unsubscribe(self, request, sso_id, *args, **kwargs):
        logger.info('ProfileCompleteUnsubscriptionApiView {}'.format(sso_id))

        requesting_user = request.user
        serializer = self.get_serializer()

        if not serializer.is_valid():
            return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)

        adapter = get_adapter(request)

        username = requesting_user.username
        email = requesting_user.email
        password = serializer.data.get('password')

        try:
            adapter.try_completely_unsubscribe(username, email, password)

        except:
            logger.exception('completely_unsubscribe error')
            response = Response('Error unsubscribing completely', status=status.HTTP_500_INTERNAL_SERVER_ERROR)

        else:
            response = Response('Successfully completely unsubscribed.', status=status.HTTP_200_OK)

        invalidate_cookie(response, app_settings.JWT_COOKIE_NAME)

        return response


class ProfileDetailApiView(generics.RetrieveAPIView):
    """
    Profile detail entrypoint
    """

    lookup_field = 'sso_id'
    serializer_class = ProfileSerializer

    def get_queryset(self):
        return Profile.objects.all()

    def get_object(self, *args, **kwargs):
        user = getattr(self.request, 'user', None)

        return user.get_sso_app_profile()
